"""
Everything here except JSONWebsockOutput class is deprecated in favor of
zorro.web
"""

import json
import abc
import logging
from http.cookies import SimpleCookie
from urllib.parse import urlparse, parse_qsl
from collections import namedtuple

import zmq

from . import redis
from .zmq import send_data
from .util import cached_property
from .di import has_dependencies, dependency


FORM_CTYPE = b'application/x-www-form-urlencoded'
log = logging.getLogger(__name__)


def blob(val):
    """Little helper function which turns data into bytes by either encoding
    or json.dumping them

    In production it can be a bad practice because you can't send json-dumped
    string (it will be just encoded). But for our example its fun that we
    can send pre-serialized data.

    And you always send json objects anyway, don't you?
    """
    if isinstance(val, (dict, list)):
        return json.dumps(val).encode('utf-8')
    elif isinstance(val, str):
        return val.encode('utf-8')
    else:
        return val


def cid(val):
    if hasattr(val, 'cid'):
        return val.cid
    assert isinstance(val, (bytes, bytearray)), ("Connection must be bytes "
        "or object having cid property")
    return val


class JSONWebsockOutput(object):

    def __init__(self, channel):
        self._channel = channel
        # we use private interface for performance
        self._sock = channel._sock

    def subscribe(self, conn, topic):
        self._do_send((b'subscribe', cid(conn), topic))

    def unsubscribe(self, conn, topic):
        self._do_send((b'unsubscribe', cid(conn), topic))

    def clone(self, source_topic, target_topic):
        self._do_send((b'clone', blob(source_topic), blob(target_topic)))

    def drop(self, topic):
        self._do_send((b'drop', topic))

    def send(self, conn, data):
        self._do_send((b'send', cid(conn), blob(data)))

    def publish(self, topic, data):
        self._do_send((b'publish', topic, blob(data)))

    def set_cookie(self, conn, cookie):
        self._do_send((b'set_cookie', cid(conn), cookie))

    def add_output(self, conn, prefix, name):
        self._do_send((b'add_output', cid(conn), prefix, name))

    def del_output(self, conn, prefix, name):
        self._do_send((b'del_output', cid(conn), prefix, name))

    def disconnect(self, conn):
        self._do_send((b'disconnect', cid(conn)))

    def _do_send(self, data):
        log.debug("Sending to zerogw: %r", data)
        # we use private interface for performance
        send_data(self._sock, data)


@has_dependencies
class JSONWebsockInput(object):

    output = dependency(JSONWebsockOutput, 'output')

    def __init__(self, prefix=''):
        self.prefix = prefix

    def __call__(self, cid, kind, *tail):
        meth = getattr(self, 'handle_' + kind.decode('ascii'), None)
        if meth:
            meth(cid, *tail)
        else:
            log.info("Unhandled message %r", kind)

    def handle_message(self, cid, body):
        parts = json.loads(body.decode('utf-8'))
        cmd = parts.pop(0)
        req_id = None
        if cmd.endswith('+'):
            cmd = cmd[:-1]
            req_id = parts.pop(0)

        if not cmd.startswith(self.prefix):
            log.info("Wrong command received %r where prefix is %r",
                     cmd, self.prefix)
            self.output.send(cid, ['_request_error', req_id, 'wrong_command'])
            return

        cmd = cmd[len(self.prefix):]
        target = getattr(self, cmd, None)
        if target is None:
            log.info("Wrong command %r", cmd)
            return
        try:
            result = self._call_convention(target, cid, *parts)
        except Exception as e:
            log.exception('Exception when processing command %r', cmd)
            if req_id:
                if hasattr(e, 'as_json'):
                    self.output.send(cid, ['_request_error',
                        req_id, e.as_json()])
                else:
                    self.output.send(cid, ['_request_error',
                        req_id, 'internal_error'])
            return
        if req_id:
            try:
                val = ['_reply', req_id,  result]
            except Exception as e:
                log.exception('Exception when serializing reply for %r', cmd)
                self.output.send(cid, ['_request_error',
                    req_id, 'cant_serialize'])
            else:
                self.output.send(cid, val)

    def _call_convention(self, target, cid, *args):
        convention = getattr(target, '__zorro_convention__', None)
        if convention == 'simple':
            return target(*args)
        elif convention == 'simple+cid':
            return target(cid, *args)
        elif convention is None:
            return
        else:
            raise UnknownConvention(
                "Unknown convention {!r}".format(convention))


class LegacyMultiDict(object):
    """Utilitary class which wrap dict to make it suitable for old utilities
    like wtforms"""

    def __init__(self, dic):
        self._dic = dic

    def getlist(self, k):
        return [self._dic[k]]

    def __contains__(self, k):
        return k in self._dic

    def __iter__(self):
        for k in self._dic:
            yield k


class RequestMixin(object):
    # TODO(tailhook) implement cookie utility
    # TODO(tailhook) implement get arguments parsing utility
    # TODO(tailhook) implement post body arguments parsing utility

    @cached_property
    def parsed_uri(self):
        return urlparse(self.uri.decode('ascii'))

    @cached_property
    def form_arguments(self):
        arguments = {}
        if hasattr(self, 'uri'):
            arguments.update(parse_qsl(self.parsed_uri.query))
        body = getattr(self, 'body', None)
        if body and getattr(self, 'content_type', None) == FORM_CTYPE:
            arguments.update(parse_qsl(self.body.decode('ascii')))
        return arguments

    @cached_property
    def legacy_arguments(self):
        return LegacyMultiDict(self.form_arguments)

    @cached_property
    def cookies(self):
        cobj = SimpleCookie(self.cookie.decode('ascii', 'ignore'))
        return dict((k, cobj[k].value) for k in cobj)


class InternalRedirect(Exception, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def update_request(self, request):
        pass


class PathRewrite(InternalRedirect):

    def __init__(self, new_path):
        self.new_path = new_path

    def update_request(self, request):
        request.current_path = self.new_path


class WebException(Exception):
    pass


class Forbidden(WebException):

    def default_response(self):
        return (b'403 Forbidden',
                b'Content-Type\0text/html\0',
                b'<!DOCTYPE html>'
                b'<html>'
                    b'<head>'
                        b'<title>403 Forbidden</title>'
                    b'</head>'
                    b'<body>'
                    b'<h1>403 Forbidden</h1>'
                    b'</body>'
                b'</html>'
                )


class InternalError(WebException):

    def default_response(self):
        return (b'500 Internal Server Error',
                b'Content-Type\0text/html\0',
                b'<!DOCTYPE html>'
                b'<html>'
                    b'<head>'
                        b'<title>500 Internal Server Error</title>'
                    b'</head>'
                    b'<body>'
                    b'<h1>500 Internal Server Error</h1>'
                    b'</body>'
                b'</html>'
                )

class NotFound(WebException):

    def default_response(self):
        return (b'404 Not Found',
                b'Content-Type\0text/html\0',
                b'<!DOCTYPE html>'
                b'<html>'
                    b'<head>'
                        b'<title>404 Page Not Found</title>'
                    b'</head>'
                    b'<body>'
                    b'<h1>404 Page Not Found</h1>'
                    b'</body>'
                b'</html>'
                )


class UnknownConvention(Exception):
    pass


class HTTPService(object):

    def __init__(self, *input_fields):
        self.Request = type('Request',
            (namedtuple('RequestBase', input_fields), RequestMixin), {})

    def _safe_dispatch(self, request):
        while True:
            try:
                result = self.dispatch(request)
            except InternalRedirect as e:
                e.update_request(request)
                continue
            except Exception as e:
                if not isinstance(e, WebException):
                    log.exception("Can't process request %r", request)
                    e = InternalError(e)
                try:
                    return self._error_page(e)
                except Exception:
                    log.exception("Can't make error page for %r", e)
                    return e.default_response()
            else:
                return result

    def _error_page(self, e):
        return e.default_response()

    def __call__(self, *args):
        request = self.Request(*args)
        result = self._safe_dispatch(request)
        if hasattr(result, 'zerogw_response'):
            return result.zerogw_response()
        else:
            return result

    def _call_convention(self, target, request):
        if target is None:
            raise NotFound()
        if hasattr(target, 'dispatch'):
            return target.dispatch(request)
        convention = getattr(target, '__zorro_convention__', None)
        if convention == 'request':
            return target(request)
        elif convention == 'simple':
            return target(request.parsed_uri.path,
                **dict(parse_qsl(request.parsed_uri.query)))
        elif convention == 'form':
            if request.form_arguments:
                args = request.legacy_arguments
                form = target.__zorro_form__(args)
                submited = True
            else:
                form = target.__zorro_form__()
                submited = False
            return target(form, submited)
        elif convention is None:
            raise Forbidden()
        else:
            raise UnknownConvention(
                "Unknown convention {!r}".format(convention))


def public_splitpath(fun):
    fun.__zorro_convention__ = 'splitpath'
    return fun


def public(fun):
    fun.__zorro_convention__ = 'simple'
    return fun


def public_with_connection_id(fun):
    fun.__zorro_convention__ = 'simple+cid'
    return fun


def public_with_request(fun):
    fun.__zorro_convention__ = 'request'
    return fun


def public_with_form(FormClass):
    def decorator(fun):
        fun.__zorro_convention__ = 'form'
        fun.__zorro_form__ = FormClass
        return fun
    return decorator


class TreeService(HTTPService):

    def __init__(self, *input_fields):
        super().__init__(*input_fields)
        self.children = {}

    def dispatch(self, request, prefix='/'):
        cpath = getattr(request, 'current_path', None)
        if cpath is None:
            cpath = request.current_path = request.parsed_uri.path

        assert cpath.startswith(prefix)
        plen = len(prefix)
        idx = cpath.find('/', plen)
        if idx < 0:
            idx = None
            next_idx = None
        else:
            next_idx = idx + 1
        part = cpath[plen:idx]
        if part in self.children:
            target = self.children[part]
        elif not part:
            target = getattr(self, 'index', None)
        elif part.startswith('_'):
            raise Forbidden()
        else:
            target = getattr(self, part, None)

        if isinstance(target, TreeService):
            return target.dispatch(request, cpath[:next_idx])
        convention = getattr(target, '__zorro_convention__', None)
        if convention == 'splitpath':
            npath = cpath[(idx or 10000)+1:]
            if npath:
                args = npath.split('/')
            else:
                args = ()
            return target(*args, **dict(parse_qsl(request.parsed_uri.query)))
        else:
            return self._call_convention(target, request)


class ParamService(HTTPService):

    def __init__(self, *input_fields):
        super().__init__(*input_fields)

    def dispatch(self, request):
        methname = request.parsed_uri.params or 'default'
        target = getattr(self, methname, None)
        return self._call_convention(target, request)


class MethodService(HTTPService):

    def __init__(self, *input_fields):
        super().__init__(*input_fields)

    def dispatch(self, request):
        methname = request.method
        target = getattr(self, methname, None)
        return self._call_convention(target, request)



"""Zorro's dependency injection framework

Usage
=====

Declare a class' dependencies::

    @has_dependencies
    class Hello(object):

        world = dependency(World, 'universe')

You must decorate class with ``@has_dependencies``. All dependencies has a
type(class), which is ``World`` in this case, and a name, which is ``universe``
in this case. This is for having multiple similar dependencies.

.. note::

    Name of dependency is unique among all services in single dependency
    injector and not tied to particular type. This is done to easier support
    subclassing of dependencies (and you can also register subclass with abc
    instead of subclassing directly)

Then at some initialisation code you create dependency injector, and set
apropriate services::

    inj = DependencyInjector()
    inj['universe'] = World()

Now you can create ``Hello`` instances and inject dependencies to them::

    hello = inj.inject(Hello())
    assert hello.world is inj['universe']

And you can propagate dependencies starting from existing instances::

    h2 = di(hello).inject(Hello())
    assert h2.world is hello.world

"""

def has_dependencies(cls):
    """Class decorator that declares dependencies"""
    deps = {}
    for i in dir(cls):
        val = getattr(cls, i)
        if isinstance(val, dependency):
            deps[i] = val
    cls.__zorro_depends__ = deps
    return cls


class dependency:
    """Property that represents single dependency"""

    def __init__(self, typ, name):
        self.type = typ
        self.name = name

    def __get__(self, inst, owner):
        if inst is None:
            return self
        raise RuntimeError("Dependency {!r} is not configured".format(self))

    def __repr__(self):
        return "<dependency {!r}:{!r}>".format(self.type, self.name)


class DependencyInjector(object):
    """Main dependency injector class

    Set all dependencies with::

        >>> inj = DependencyInjector()
        >>> inj['name'] = value
        >>> obj = inj.inject(DependentClass())

    Propagate dependencies with::

        >>> di(obj).inject(NewObject())
    """

    def __init__(self):
        self._provides = {}

    def inject(self, inst, **renames):
        """Injects dependencies and propagates dependency injector"""
        if renames:
            di = self.clone(**renames)
        else:
            di = self
        pro = di._provides
        inst.__zorro_di__ = di
        deps = getattr(inst, '__zorro_depends__', None)
        if deps:
            for attr, dep in deps.items():
                val = pro[dep.name]
                if not isinstance(val, dep.type):
                    raise RuntimeError("Wrong provider for {!r}".format(val))
                setattr(inst, attr, val)
        meth = getattr(inst, '__zorro_di_done__', None)
        if meth is not None:
            meth()
        return inst

    def inject_subset(self, inst, *names, **renames):
        """Injects only specified dependencies, and does not propagate others
        """
        # TODO(tailhook) may be optimize if names == _provides.keys()
        di = self.__class__()
        mypro = self._provides
        pro = di._provides
        for name in names:
            pro[name] = mypro[name]
        for name, alias in renames.items():
            pro[name] = mypro[alias]
        inst.__zorro_di__ = di
        deps = getattr(inst, '__zorro_depends__', None)
        if deps:
            for attr, dep in deps.items():
                if dep.name not in pro:
                    continue
                val = pro[dep.name]
                if not isinstance(val, dep.type):
                    raise RuntimeError("Wrong provider for {!r}".format(val))
                setattr(inst, attr, val)
        meth = getattr(inst, '__zorro_di_done__', None)
        if meth is not None:
            meth()
        return inst

    def inject_restricted(self, inst):
        """Injects dependencies, and propagates only this instance's
        dependencies
        """
        # TODO(tailhook) may be optimize if names == _provides.keys()
        di = self.__class__()
        mypro = self._provides
        pro = di._provides
        for name in names:
            pro[name] = mypro[name]
        for name, alias in renames.items():
            pro[name] = mypro[alias]
        inst.__zorro_di__ = di
        deps = getattr(inst, '__zorro_depends__', None)
        if deps:
            for attr, dep in deps.items():
                val = pro[dep.name]
                if not isinstance(val, dep.type):
                    raise RuntimeError("Wrong provider for {!r}".format(val))
                setattr(inst, attr, val)
        meth = getattr(inst, '__zorro_di_done__', None)
        if meth is not None:
            meth()
        return inst

    def __setitem__(self, name, value):
        if name in self._provides:
            raise RuntimeError("Two providers for {!r}".format(name))
        self._provides[name] = value

    def __getitem__(self, name):
        return self._provides[name]

    def __contains__(self, name):
        return name in self._provides

    def clone(self, **renames):
        di = self.__class__()
        mypro = self._provides
        pro = di._provides
        pro.update(mypro)
        for name, alias in renames.items():
            pro[name] = mypro[alias]
        return di


def dependencies(cls):
    """Returns dict of dependencies of a class declared with
    ``@has_dependencies``
    """
    return getattr(cls, '__zorro_depends__', {})


def di(obj):
    """Returns dependency injector used to construct class

    This instance is useful to propagate dependencies
    """
    try:
        return obj.__zorro_di__
    except AttributeError:
        raise RuntimeError("No dependency injector found")



import struct
from binascii import hexlify
from itertools import count

_unpack_dict = {}
_pack_dict = {}


class ObjectID(bytes):

    def __repr__(self):
        return 'ObjectID.fromhex("{}")'.format(hexlify(self).decode('ascii'))


def _unpack(char):
    def register(fun):
        _unpack_dict[char] = fun
        return fun
    return register


def _pack(char, typ):
    def register(fun):
        _pack_dict[typ] = (char, fun)
        return fun
    return register


@_unpack(0x01)
def unpack_float(buf, idx):
    return struct.unpack_from('<d', buf, idx)[0], idx + 8


@_unpack(0x02)
def unpack_string(buf, idx):
    ln, = struct.unpack_from('<i', buf, idx)
    idx += 4
    if buf[idx+ln-1] != 0:
        raise ValueError("String serialization is broken")
    return buf[idx:idx+ln-1].decode('utf-8'), idx+ln


@_unpack(0x03)
def unpack_document(buf, idx):
    doc_len, = struct.unpack_from('<i', buf, idx)
    if buf[idx+doc_len-1]:
        raise ValueError("Wrong trailing byte")
    idx += 4
    obj = {}
    while buf[idx]:
        typ = buf[idx]
        nameend = buf.index(b'\x00', idx+1)
        fun = _unpack_dict.get(typ)
        if fun is None:
            raise ValueError("Wrong value type")
        obj[buf[idx+1:nameend].decode('utf-8')], idx = fun(buf, nameend+1)
    return obj, idx+1


@_unpack(0x04)
def unpack_array(buf, idx):
    doc_len, = struct.unpack_from('<i', buf, idx)
    if buf[idx+doc_len-1]:
        raise ValueError("Wrong trailing byte")
    idx += 4
    obj = []
    for i in count():
        typ = buf[idx]
        if not typ:
            break
        nameend = buf.index(b'\x00', idx+1)
        fun = _unpack_dict.get(typ)
        if fun is None:
            raise ValueError("Wrong value type")
        if int(buf[idx+1:nameend]) != i:
            raise ValueError("Wrong array index")
        val, idx = fun(buf, nameend+1)
        obj.append(val)
    return obj, idx+1


@_unpack(0x07)
def unpack_objectid(buf, idx):
    return ObjectID(buf[idx:idx+12]), idx+12


@_unpack(0x0A)
def unpack_none(buf, idx):
    return None, idx


@_unpack(0x10)
def unpack_int32(buf, idx):
    return struct.unpack_from('<i', buf, idx)[0], idx + 4


@_unpack(0x12)
def unpack_int64(buf, idx):
    return struct.unpack_from('<q', buf, idx)[0], idx + 8


def loads(s):
    if len(s) < 5:
        raise ValueError("Too short string")
    obj, idx = unpack_document(s, 0)
    if idx != len(s):
        if idx < len(s):
            raise ValueError("Premature end of document")
        else:
            raise ValueError("Garbage at end of document")
    return obj


def iter_load_from(buf, offset=0):
    while offset < len(buf):
        obj, offset = unpack_document(buf, offset)
        yield obj


@_pack(0x01, float)
def pack_float(value, buf):
    buf.extend(struct.pack('<d', value))


@_pack(0x02, str)
def pack_string(value, buf):
    val = value.encode('utf-8')
    buf.extend(struct.pack('<i', len(val)+1))
    buf.extend(val)
    buf.append(0)


@_pack(0x03, dict)
def pack_document(value, buf):
    pos = len(buf)
    buf += b'\x00\x00\x00\x00'
    _pack_doc(value.items(), buf)
    buf.append(0)
    struct.pack_into('<i', buf, pos, len(buf) - pos)


@_pack(0x04, list)
def pack_array(value, buf):
    pos = len(buf)
    buf += b'\x00\x00\x00\x00'
    _pack_doc(((str(i), v)
               for i, v in enumerate(value)), buf)
    buf.append(0)
    struct.pack_into('<i', buf, pos, len(buf) - pos)


@_pack(0x07, ObjectID)
def pack_objectid(value, buf):
    assert len(value) == 12, "ObjectID length is not 12"
    buf += value


@_pack(0x0A, type(None))
def pack_none(value, buf):
    """Nothing needed to pack None, type byte is in another place"""


@_pack(0x10, int)
def pack_int(value, buf):
    if value > (1 << 31) or value < -(1 << 31):
        raise NotImplementedError("Long integers not implemented")
    buf.extend(struct.pack('<i', value))


def _pack_doc(pairs, buf):
    for k, v in pairs:
        spec = _pack_dict.get(type(v))
        if spec is None:
            for typ in _pack_dict:
                if isinstance(v, typ):
                    spec = _pack_dict[typ]
                    break
            else:
                raise ValueError("Can't BSONize {}".format(type(v)))
        ch, fun = spec
        buf.append(ch)
        buf.extend(k.encode('utf-8'))
        buf.append(0)
        fun(v, buf)


def dumps(obj):
    buf = bytearray(4)
    _pack_doc(obj.items(), buf)
    buf.append(0)
    struct.pack_into('<i', buf, 0, len(buf))
    return bytes(buf)


def dump_extend(buf, obj):
    pos = len(buf)
    buf += b'\x00\x00\x00\x00'
    _pack_doc(obj.items(), buf)
    buf.append(0)
    struct.pack_into('<i', buf, pos, len(buf)-pos)


def dump_extend_iter(buf, iterable):
    pos = len(buf)
    buf += b'\x00\x00\x00\x00'
    _pack_doc(iterable, buf)
    buf.append(0)
    struct.pack_into('<i', buf, pos, len(buf) - pos)



import socket
import errno

from .core import gethub, Lock
from . import channel
from .util import setcloexec


convert = {
    str: lambda a: a.encode('utf-8'),
    bytes: lambda a: a,
    bytearray: lambda a: a,
    int: lambda a: bytes(str(a), 'utf-8'),
    float: lambda a: bytes(repr(a), 'utf-8'),
    }


def encode_command(buf, parts):
    add = buf.extend
    cvt = convert
    add('*{:d}\r\n'.format(len(parts)).encode('ascii'))
    for part in parts:
        value = cvt[part.__class__](part)
        add('${:d}\r\n'.format(len(value)).encode("ascii"))
        add(value)
        add(b'\r\n')
    return buf


class RedisError(Exception):
    pass


class RedisChannel(channel.PipelinedReqChannel):
    BUFSIZE = 16384

    def __init__(self, host, port, unixsock, db):
        super().__init__()
        if unixsock:
            self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        else:
            self._sock = socket.socket(socket.AF_INET,
                socket.SOCK_STREAM, socket.IPPROTO_TCP)
        setcloexec(self._sock)
        self._sock.setblocking(0)
        try:
            if unixsock:
                self._sock.connect(unixsock)
            else:
                self._sock.connect((host, port))
        except socket.error as e:
            if e.errno == errno.EINPROGRESS:
                gethub().do_write(self._sock)
            else:
                raise
        self._start()
        db = str(db)
        assert self.request('*2\r\n$6\r\nSELECT\r\n${0}\r\n{1}\r\n'
            .format(len(db), db).encode('ascii')).get() == 'OK'

    def _close_channel(self):
        self._sock.close()
        super()._close_channel()

    def sender(self):
        buf = bytearray()

        add_chunk = buf.extend
        wait_write = gethub().do_write

        while True:
            if not buf:
                self.wait_requests()
            if not self._alive:
                return
            wait_write(self._sock)
            for chunk in self.get_pending_requests():
                add_chunk(chunk)
            try:
                bytes = self._sock.send(buf)
            except socket.error as e:
                if e.errno in (errno.EAGAIN, errno.EINTR):
                    continue
                else:
                    raise
            if not bytes:
                raise EOFError()
            del buf[:bytes]

    def receiver(self):
        buf = bytearray()

        sock = self._sock
        wait_read = gethub().do_read
        add_chunk = buf.extend
        pos = [0]

        def readmore():
            while True:
                wait_read(sock)
                try:
                    if pos[0]*2 > len(buf):
                        del buf[:pos[0]]
                        pos[0] = 0
                    bytes = sock.recv(self.BUFSIZE)
                    if not bytes:
                        raise EOFError()
                    add_chunk(bytes)
                except socket.error as e:
                    if e.errno in (errno.EAGAIN, errno.EINTR):
                        continue
                    else:
                        raise
                else:
                    break

        def readchar():
            if len(buf) <= pos[0]:
                readmore()
            c = buf[pos[0]]
            pos[0] += 1
            return c

        def readline():
            if len(buf) < 2 or pos[0] >= len(buf):
                readmore()
            while True:
                try:
                    idx = buf.index(b'\r\n', pos[0])
                except ValueError:
                    pass
                else:
                    break
                readmore()
            res = buf[pos[0]:idx]
            pos[0] = idx + 2
            return res

        def readslice(ln):
            while len(buf) - pos[0] < ln:
                readmore()
            res = buf[pos[0]:pos[0]+ln]
            pos[0] += ln
            return res

        def readone():
            ch = readchar()
            if ch == 42: # b'*'
                cnt = int(readline())
                return [readone() for i in range(cnt)]
            elif ch == 43: # b'+'
                return readline().decode('ascii')
            elif ch == 45: # b'-'
                return RedisError(readline().decode('ascii'))
            elif ch == 58: # b':'
                return int(readline())
            elif ch == 36: # b'$'
                ln = int(readline())
                if ln < 0:
                    return None
                res = readslice(ln)
                assert readline() == b''
                return res
            else:
                raise NotImplementedError(ch)

        while True:
            self.produce(readone())


class Redis(object):
    def __init__(self, host='localhost', port=6379, unixsock=None, db=0):
        self.unixsock = unixsock
        self.host = host
        self.port = port
        self.db = db
        self._channel = None
        self._channel_lock = Lock()

    def check_connection(self):
        if not self._channel:
            with self._channel_lock:
                if not self._channel:
                    self._channel = RedisChannel(self.host, self.port,
                        db=self.db, unixsock=self.unixsock)

    # low-level stuff
    def execute(self, *args):
        self.check_connection()
        buf = bytearray()
        encode_command(buf, args)
        return self._channel.request(buf).get()

    def future(self, *args):
        self.check_connection()
        buf = bytearray()
        encode_command(buf, args)
        return self._channel.request(buf)

    def pipeline(self, commands):
        self.check_connection()
        buf = bytearray()
        for cmd in commands:
            encode_command(buf, cmd)
        return self._channel.request(buf, len(commands)).get()

    def bulk(self, commands):
        self.check_connection()
        if commands[0][0] != 'MULTI' or commands[-1][0] != 'EXEC':
            raise ValueError("Bulk must start with MULTI and end with EXEC")
        buf = bytearray()
        for cmd in commands:
            encode_command(buf, cmd)
        val = self._channel.request(buf, len(commands)).get()
        if val[0] != 'OK':
            raise RuntimeError(val, commands)
        for i in val[1:-1]:
            if isinstance(i, RedisError):
                raise i
            assert i == 'QUEUED'
        return val[-1]
    # high-level stuff



import fcntl
from heapq import heappush, heappop
from collections import deque


class orderedset(object):

    def __init__(self):
        self.deque = deque()

    def add(self, *val):
        self.deque.append(val)
        def remover(*a):
            self.deque.remove(val)
        return remover

    def update(self, value):
        self.deque.extend(value)

    def first(self):
        if self.deque:
            return self.deque[0]
        return None

    def remove(self, value):
        self.deque.remove(value)

    def __bool__(self):
        return bool(self.deque)


class priorityqueue(object):

    def __init__(self):
        self.heap = []
        self.counter = 0

    def add(self, pri, task):
        self.counter += 1
        item = [pri, self.counter, task]
        heappush(self.heap, item)
        def remover(*a):
            item[2] = None
        return remover

    def min(self):
        while self.heap and self.heap[0][2] is None:
            heappop(self.heap)
        if self.heap:
            return self.heap[0][0]
        return None

    def pop(self, value):
        val = self.min()
        if val is not None and val <= value:
            return self.heap[0][2]
        return None

    def __bool__(self):
        return bool(self.heap)


try:
    from socket import socketpair
except ImportError:
    import socket
    def socketpair():
        s=socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(('127.0.0.1', 0))
        host, port = s.getsockname()
        s.listen(1)
        a=socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
        a.connect((host, port))
        b, _ = s.accept()
        s.close()
        a.setblocking(0)
        b.setblocking(0)
        return a, b


def socket_pair():
    a, b = socketpair()
    a.setblocking(0)
    b.setblocking(0)
    return a, b


class cached_property(object):

    def __init__(self, fun):
        self.function = fun
        self.name = fun.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        res = obj.__dict__[self.name] = self.function(obj)
        return res


class marker_object(object):
    __slots__ = ('name',)

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '<{}>'.format(self.name)


def setcloexec(sock):
    flags = fcntl.fcntl(sock, fcntl.F_GETFD)
    fcntl.fcntl(sock, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)



from collections import deque
import logging

from greenlet import GreenletExit

from . import Future, Condition, gethub


log = logging.getLogger(__name__)


class PipeError(Exception):
    pass


class ShutdownException(Exception):
    pass


class BaseChannel(object):
    def __init__(self):
        self._alive = True
        self._pending = deque()
        self._cond = Condition()
        self._sender_alive = True
        self._receiver_alive = True

    def _start(self):
        hub = gethub()
        hub.do_spawnhelper(self._sender_wrapper)
        hub.do_spawnhelper(self._receiver_wrapper)

    def __bool__(self):
        return self._alive

    def _close_channel(self):
        pass

    def _sender_wrapper(self):
        try:
            self.sender()
        except (EOFError, ShutdownException, GreenletExit) as e:
            pass
        except Exception:
            log.exception("Error in %r's sender", self)
        finally:
            if self._alive:
                self._alive = False
                self._stop_producing()
            self._sender_alive = False
            if not self._receiver_alive:
                # both are down
                self._close_channel()

    def _receiver_wrapper(self):
        try:
            self.receiver()
        except (EOFError, ShutdownException, GreenletExit) as e:
            pass
        except Exception:
            log.exception("Error in %r's receiver", self)
        finally:
            if self._alive:
                self._alive = False
                self._stop_producing()
            self._receiver_alive = False
            if not self._sender_alive:
                # both are down
                self._close_channel()

    def _stop_producing(self):
        pass

    def peek_request(self):
        self.wait_requests()
        return self._pending[0]

    def pop_request(self):
        return self._pending.popleft()

    def wait_requests(self):
        while not self._pending:
            if not self._alive:
                raise ShutdownException()
            self._cond.wait()

    def get_pending_requests(self):
        while self._pending:
            yield self._pending.popleft()


class PipelinedReqChannel(BaseChannel):

    def __init__(self):
        super().__init__()
        self._producing = deque()
        self._cur_producing = []

    def _stop_producing(self):
        prod = self._producing
        del self._producing
        for num, fut in prod:
            fut.throw(PipeError())
        self._cond.notify()  # wake up consumer if it waits for messages

    def produce(self, value):
        if not self._alive:
            raise ShutdownException()
        val = self._cur_producing.append(value)
        num = self._producing[0][0]
        if num is None:
            del self._cur_producing[:]
            self._producing.popleft()[1].set(value)
        elif len(self._cur_producing) >= num:
            res = tuple(self._cur_producing)
            del self._cur_producing[:]
            self._producing.popleft()[1].set(res)
        else:
            return

    def request(self, input, num_output=None):
        if not self._alive:
            raise PipeError()
        val = Future()
        self._pending.append(input)
        self._producing.append((num_output, val))
        self._cond.notify()
        return val

    def push(self, input):
        """For requests which do not need an answer"""
        self._pending.append(input)
        self._cond.notify()


class MuxReqChannel(BaseChannel):
    def __init__(self):
        super().__init__()
        self.requests = {}

    def new_id(self):
        raise NotImplementedError("Abstract method")

    def _stop_producing(self):
        reqs = self.requests
        del self.requests
        for fut in reqs.values():
            fut.throw(PipeError())

    def request(self, input):
        if not self._alive:
            raise PipeError()
        id = self.new_id()
        assert not id in self.requests
        val = Future()
        self.requests[id] = val
        self._pending.append((id, input))
        self._cond.notify()
        return val

    def push(self, input):
        """For requests which do not need an answer"""
        id = self.new_id()
        assert not id in self.requests
        self._pending.append((id, input))
        self._cond.notify()

    def produce(self, id, data):
        if not self._alive:
            raise ShutdownException()
        fut = self.requests.pop(id, None)
        if fut is not None:
            fut.set(data)




import os
import random
import struct
import errno
import time
import pickle
from functools import partial

import zmq
from zmq import *

from . import core, channel
from .core import gethub

try:
    REP_SOCKET = zmq.XREP
except AttributeError:
    REP_SOCKET = zmq.ROUTER
try:
    REQ_SOCKET = zmq.XREQ
except AttributeError:
    REQ_SOCKET = zmq.DEALER

DEFAULT_IO_THREADS = 1


class Socket(zmq.Socket):

    def dict_configure(self, data):
        if 'bind' in data:
            if isinstance(data['bind'], (tuple, list)):
                for addr in data['bind']:
                    self.bind(addr)
            else:
                self.bind(data['bind'])
        if 'connect' in data:
            if isinstance(data['connect'], (tuple, list)):
                for addr in data['connect']:
                    self.connect(addr)
            else:
                self.connect(data['connect'])
        # we have no identity as swap as they will be dropped
        if 'hwm' in data:
            self.setsockopt(zmq.HWM, int(data['hwm']))
        if 'affinity' in data:
            self.setsockopt(zmq.AFFINITY, int(data['affinity']))
        if 'backlog' in data:
            self.setsockopt(zmq.BACKLOG, int(data['backlog']))
        if 'linger' in data:
            self.setsockopt(zmq.LINGER, int(data['linger']))
        if 'sndbuf' in data:
            self.setsockopt(zmq.SNDBUF, int(data['sndbuf']))
        if 'rcvbuf' in data:
            self.setsockopt(zmq.RCVBUF, int(data['rcvbuf']))
        # TODO(tailhook) add other options


def send_data(sock, data, address=None):
    if address is None:
        _rep = []
    else:
        _rep = list(address)
    for a in data:
        if isinstance(a, str):
            _rep.append(a.encode('utf-8'))
        else:
            _rep.append(a)
    while True:
        try:
            sock.send_multipart(_rep, zmq.NOBLOCK)
        except zmq.ZMQError as e:
            if e.errno == errno.EAGAIN:
                gethub().do_write(sock)
                continue
            elif e.errno == errno.EINTR:
                continue
            else:
                raise
        else:
            break


def rep_responder(sock, address, callback, data):
    hub = core.gethub()
    reply = callback(*data)
    if isinstance(reply, bytes):
        send_data(sock, (reply,), address=address)
    elif isinstance(reply, str):
        send_data(sock, (reply.encode('utf-8'),), address=address)
    elif reply is None:
        raise RuntimeError("Replier callback must return either string,"
            " bytes or sequence of strings or bytes")
    else:
        send_data(sock, reply, address=address)


def rep_listener(sock, callback):
    hub = core.gethub()
    # hook for pools
    wait_slot = getattr(callback, 'wait_slot', None)
    while True:
        hub.do_read(sock)
        while True:
            if wait_slot is not None:
                wait_slot()
            try:
                data = sock.recv_multipart(zmq.NOBLOCK)
            except zmq.ZMQError as e:
                if e.errno == errno.EAGAIN or e.errno == errno.EINTR:
                    break
                else:
                    raise
            i = data.index(b'')
            addr = data[:i+1]
            data = data[i+1:]
            # we create new greenlet and immediately switch to it
            # this effectively starts processing request faster
            # but more importantly counts processing request in request pool
            hub.do_spawnswitch(
                partial(rep_responder, sock, addr, callback, data))


def rep_socket(callback):
    sock = Socket(context(), REP_SOCKET)
    core.gethub().do_spawnservice(partial(rep_listener, sock, callback))
    return sock


def sub_listener(sock, callback):
    hub = core.gethub()
    # hooks for pools
    wait_slot = getattr(callback, 'wait_slot', None)
    while True:
        hub.do_read(sock)
        while True:
            if wait_slot is not None:
                wait_slot()
            try:
                data = sock.recv_multipart(zmq.NOBLOCK)
            except zmq.ZMQError as e:
                if e.errno == errno.EAGAIN or e.errno == errno.EINTR:
                    break
                else:
                    raise
            # we create new greenlet and immediately switch to it
            # this effectively starts processing request faster
            # but more importantly counts processing request in request pool
            hub.do_spawnswitch(partial(callback, *data))


def sub_socket(callback):
    sock = Socket(context(), zmq.SUB)
    core.gethub().do_spawnservice(partial(sub_listener, sock, callback))
    return sock


def pub_socket():
    return PubChannel()


def pull_socket(callback):
    sock = Socket(context(), zmq.PULL)
    core.gethub().do_spawnservice(partial(sub_listener, sock, callback))
    return sock


def push_socket():
    return PushChannel()


def req_socket():
    return ReqChannel()


class ReqChannel(channel.MuxReqChannel):

    def __init__(self):
        super().__init__()
        self.init_id()
        self._sock = Socket(context(), REQ_SOCKET)
        self._start()

    def init_id(self):
        self.prefix = struct.pack('HHL',
            os.getpid() % 65536, random.randrange(65536), int(time.time()))
        self.counter = 0

    def new_id(self):
        self.counter += 1
        if self.counter >= 4294967296:
            self.init_id()
        return self.prefix + struct.pack('L', self.counter)

    def bind(self, value):
        self._sock.bind(value)

    def connect(self, value):
        self._sock.connect(value)

    def dict_configure(self, dic):
        self._sock.dict_configure(dic)

    def sender(self):
        wait_write = core.gethub().do_write
        while True:
            wait_write(self._sock)
            id, data = self.peek_request()
            self._sock.send(id, zmq.SNDMORE)
            self._sock.send(b"", zmq.SNDMORE)
            self._sock.send_multipart(data)
            self.pop_request()

    def receiver(self):
        wait_read = core.gethub().do_read
        while True:
            wait_read(self._sock)
            data = self._sock.recv_multipart()
            assert data[1] == b''
            self.produce(data[0], data[2:])


class OutputChannel(object):
    zmq_kind = None

    def __init__(self):
        super().__init__()
        self._sock = Socket(context(), self.zmq_kind)

    def bind(self, value):
        self._sock.bind(value)

    def connect(self, value):
        self._sock.connect(value)

    def dict_configure(self, dic):
        self._sock.dict_configure(dic)


class PubChannel(OutputChannel):
    zmq_kind = zmq.PUB

    def publish(self, *args):
        send_data(self._sock, args)


class PushChannel(OutputChannel):
    zmq_kind = zmq.PUSH

    def push(self, *args):
        send_data(self._sock, args)


def _get_fd(value):
    if isinstance(value, zmq.Socket):
        return value
    else:
        return value.fileno()


def plug(hub, io_threads=DEFAULT_IO_THREADS):
    assert not hasattr(hub, 'zmq_context')
    ctx = hub.zmq_context = zmq.Context(io_threads)
    hub.change_poller(zmq.Poller, filedes=_get_fd,
        POLLIN=zmq.POLLIN, POLLOUT=zmq.POLLOUT, POLLERR=zmq.POLLERR, POLLHUP=0)
    core.os_errors += (ZMQError,)
    hub.log_plugged(ctx, name='zmq_context')
    return ctx


def context():
    hub = core.gethub()
    ctx = getattr(hub, 'zmq_context', None)
    if ctx is None:
        return plug(hub)
    return ctx


class MethodCallError(Exception):
    pass


class MethodException(Exception):
    pass


class Method(object):
    __slots__ = ('owner', 'name')

    def __init__(self, owner, name):
        self.owner = owner
        self.name = name

    def __call__(self, *args):
        o = self.owner
        lst = [self.name]
        lst.extend(map(o.dumps, args))
        res = o.channel.request(lst).get()
        if res[0] == b'_result':
            return self.owner.loads(res[1])
        elif res[0] == b'_error':
            raise MethodCallError(res[1].decode('ascii'))
        elif res[0] == b'_exception':
            raise MethodException(res[1].decode('utf-8'))
        else:
            raise MethodException("Wrong reply")


class Requester(object):
    loads = pickle.loads
    dumps = pickle.dumps

    def __init__(self, channel, prefix=''):
        self.channel = channel
        if isinstance(prefix, str):
            self.prefix = prefix.encode('ascii')

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError("Method name can't start with underscore")
        fullname = self.prefix + name.encode('ascii')
        return Method(self, fullname)


class Responder(object):
    loads = pickle.loads
    dumps = pickle.dumps

    def __call__(self, name, *args):
        name = name.decode('ascii')
        if name.startswith('_'):
            return (b'_error', b'bad_name')
        meth = getattr(self, name, None)
        if meth is None or not callable(meth):
            return (b'_error', b'no_method')
        try:
            args = tuple(map(self.loads, args))
        except Exception as e:
            return (b'_error', b'unpacking_error')
        try:
            result = meth(*args)
        except Exception as e:
            # TODO(tailhook) log exception
            return (b'_exception', repr(e))
        try:
            result = self.dumps(result)
        except Exception as e:
            return (b'_error', b'packing_error')
        return (b'_result', result)


class Dispatcher(dict):

    def __init__(self, default, **kw):
        if isinstance(default, dict):
            super().__init__(default, **kw)
        else:
            self[None] = default
            super().__init__(**kw)

    def __call__(self, name, *args):
        parts = name.rsplit(b'.', 1)
        if len(parts) > 1:
            oname, name = parts
            oname = oname.decode('ascii')
        else:
            oname = None
            name = parts[0]
        obj = self.get(oname)
        if obj is None:
            return (b'_error', b'wrong_prefix')
        else:
            return obj(name, *args)





import socket
import errno
import struct
import os.path

from ..core import gethub, Lock
from .. import channel
from . import bson
from ..util import setcloexec


OP_REPLY = 1
OP_MSG = 1000
OP_UPDATE = 2001
OP_INSERT = 2002
OP_QUERY = 2004
OP_GET_MORE	= 2005
OP_DELETE = 2006
OP_KILL_CURSORS	= 2007


class MongodbError(Exception):
    pass


class Channel(channel.MuxReqChannel):
    BUFSIZE = 16384

    def __init__(self, host, port, socket_dir='/tmp'):
        super().__init__()
        sock = None
        if host in ('localhost', '127.0.0.1'):
            unix_sock = os.path.join(socket_dir,
                'mongodb-{}.sock'.format(port))
            if os.path.exists(unix_sock):
                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                addr = unix_sock
        if sock is None:
            sock = socket.socket(socket.AF_INET,
                socket.SOCK_STREAM, socket.IPPROTO_TCP)
            addr = (host, port)
        setcloexec(sock)
        self._sock = sock
        self._sock.setblocking(0)
        try:
            self._sock.connect(addr)
        except socket.error as e:
            if e.errno == errno.EINPROGRESS:
                gethub().do_write(self._sock)
            else:
                raise
        self._counter = 0
        self._start()

    def new_id(self):
        self._counter += 1
        if self._counter >= 1 << 31:
            self._counter -= 1 << 31
        return self._counter

    def sender(self):
        buf = bytearray()

        add_chunk = buf.extend
        wait_write = gethub().do_write
        sock = self._sock

        while True:
            if not buf:
                self.wait_requests()
            wait_write(sock)
            for id, chunk in self.get_pending_requests():
                struct.pack_into('<ii', chunk, 0, len(chunk), id)
                add_chunk(chunk)
            try:
                bytes = sock.send(buf)
            except socket.error as e:
                if e.errno in (errno.EAGAIN, errno.EINTR):
                    continue
                else:
                    raise
            if not bytes:
                raise EOFError()
            del buf[:bytes]

    def receiver(self):
        buf = bytearray()

        sock = self._sock
        wait_read = gethub().do_read
        add_chunk = buf.extend
        pos = 0

        while True:
            if pos*2 > len(buf):
                del buf[:pos]
                pos = 0
            wait_read(sock)
            try:
                bytes = sock.recv(self.BUFSIZE)
                if not bytes:
                    raise EOFError()
                add_chunk(bytes)
            except socket.error as e:
                if e.errno in (errno.EAGAIN, errno.EINTR):
                    continue
                else:
                    raise
            while len(buf)-pos >= 16:
                length, resp_to = struct.unpack_from('<i4xi', buf, pos)
                if len(buf)-pos < length:
                    break
                self.produce(resp_to, buf[pos:pos+length])
                pos += length


class Connection(object):

    def __init__(self, host='127.0.0.1', port=27017, socket_dir='/tmp'):
        self._channel = None
        self._channel_lock = Lock()
        self._databases = {}
        self.host = host
        self.port = port
        self.socket_dir = socket_dir

    def __getitem__(self, db_name):
        try:
            return self._databases[db_name]
        except KeyError:
            res = Database(db_name, self)
            self._databases[db_name] = res
            return res

    def channel(self):
        if not self._channel:
            with self._channel_lock:
                if not self._channel:
                    self._channel = Channel(
                        self.host, self.port,
                        socket_dir=self.socket_dir)
        return self._channel


class Database(object):

    def __init__(self, name, conn):
        self._name = name
        self._conn = conn
        self._collections = {}

    def __getitem__(self, coll_name):
        try:
            return self._collections[coll_name]
        except KeyError:
            res = Collection(self._name, coll_name, self._conn)
            self._collections[coll_name] = res
            return res


class Collection(object):

    def __init__(self, db_name, name, conn):
        self._name = name
        self._db_name = db_name
        self._fullname_cs = ('{}.{}\0'
            .format(self._db_name, self._name)
            .encode('ascii'))
        self._conn = conn

    def insert(self, record):
        flags = 0  # single record, no need to continue on error
        req = bytearray(struct.pack('<12xii', OP_INSERT, flags))
        req += self._fullname_cs
        bson.dump_extend(req, record)
        self._conn.channel().push(req)

    def insert_many(self, records, continue_on_error=True):
        flags = 1 if continue_on_error else 0
        req = bytearray(struct.pack('<12xii', OP_INSERT, flags))
        req += self._fullname_cs
        for record in records:
            bson.dump_extend(req, record)
        self._conn.channel().push(req)

    def query(self, query, *,
        fields=None, skip=0, limit=0,
        slave_ok=False, partial_ok=False):
        flags = 0
        if slave_ok:
            flags |= 1 << 2
        if partial_ok:
            flags |= 1 << 7
        req = bytearray(struct.pack('<12xii', OP_QUERY, flags))
        req += self._fullname_cs
        req += struct.pack('ii', skip, limit)
        bson.dump_extend(req, query)
        if fields is not None:
            if isinstance(fields, (set, frozenset, list, tuple)):
                bson.dump_extend_iter(req, ((k, 1) for k in fields))
            else:
                bson.dump_extend(req, fields)
        res = self._conn.channel().request(req).get()
        op_code, flags, _, start, num = struct.unpack_from('<12xiiqii', res, 0)
        if op_code != OP_REPLY:
            raise MongodbError("Wrong kind of reply returned")
        if start != skip:
            raise MongodbError("Mongodb returned wrong start number")
        docs = bson.iter_load_from(res, 36)  # struct.calcsize('<12xiiqii')
        if flags & (1 << 1):
            raise MongodbError(next(iter(docs))['$err'])
        return docs

    def delete(self, selector, *, single=False):
        flags = 1 if single else 0
        if not selector:
            raise ValueError("Empty selector for delete() use clean()")
        req = bytearray(struct.pack('<12xi4x', OP_DELETE))
        req += self._fullname_cs
        req += struct.pack('<i', flags)
        bson.dump_extend(req, selector)
        self._conn.channel().push(req)

    def clean(self):
        req = bytearray(struct.pack('<12xi4x', OP_DELETE))
        req += self._fullname_cs
        req += b'\x00\x00\x00\x00\x05\x00\x00\x00\x00'
        self._conn.channel().push(req)

    def update(self, selector, data, *, upsert=False, multi=False):
        flags = 0
        if upsert:
            flags |= 1 << 0
        if multi:
            flags |= 1 << 1
        req = bytearray(struct.pack('<12xi4x', OP_UPDATE))
        req += self._fullname_cs
        req += struct.pack('<i', flags)
        bson.dump_extend(req, selector)
        bson.dump_extend(req, data)
        self._conn.channel().push(req)

    def save(self, doc):
        self.update({ '_id': doc['_id'] }, doc, upsert=True)






import abc
import json
import logging
import inspect
from http.cookies import SimpleCookie
from urllib.parse import urlparse, parse_qsl
from itertools import zip_longest
from functools import partial

from .util import cached_property, marker_object


log = logging.getLogger(__name__)

_LEAF_METHOD = marker_object('LEAF_METHOD')
_LEAF_WSOCK_METHOD = marker_object('LEAF_WSOCK_METHOD')
_LEAF_HTTP_METHOD = marker_object('LEAF_HTTP_METHOD')
_RESOURCE_METHOD = marker_object('RESOURCE_METHOD')
_RES_WSOCK_METHOD = marker_object('RES_WSOCK_METHOD')
_RES_HTTP_METHOD = marker_object('RES_HTTP_METHOD')
_RESOURCE = marker_object('RESOURCE')
_INTERRUPT = marker_object('INTERRUPT')
_FORM_CTYPE = b'application/x-www-form-urlencoded'


class LegacyMultiDict(object):
    """Utilitary class which wrap dict to make it suitable for old utilities
    like wtforms"""

    def __init__(self, pairs=None):
        self._dic = {}
        if not pairs is None:
            self.update(pairs)

    def update(self, pairs):
        for k, v in pairs:
            if k not in self._dic:
                self._dic[k] = [v]
            else:
                self._dic[k].append(v)

    def getlist(self, k):
        return list(self._dic[k])

    def __contains__(self, k):
        return k in self._dic

    def __iter__(self):
        for k in self._dic:
            yield k

    def __len__(self):
        return len(self._dic)


class Request(object):
    __slots__ = ('__dict__',
        # the excpected property names and names of respective zerogw items
        'uri',           # !Uri
        'content_type',  # !Header Content-Type
        'cookie',        # !Header Cookie
        'body',          # !PostBody
        )

    @cached_property
    def parsed_uri(self):
        return urlparse(self.uri.decode('ascii'))

    @cached_property
    def form_arguments(self):
        arguments = {}
        if hasattr(self, 'uri'):
            arguments.update(parse_qsl(self.parsed_uri.query))
        body = getattr(self, 'body', None)
        if body and getattr(self, 'content_type', None) == _FORM_CTYPE:
            arguments.update(parse_qsl(self.body.decode('ascii')))
        return arguments

    @cached_property
    def legacy_arguments(self):
        arguments = LegacyMultiDict()
        if hasattr(self, 'uri'):
            arguments.update(parse_qsl(self.parsed_uri.query))
        body = getattr(self, 'body', None)
        if body and getattr(self, 'content_type', None) == _FORM_CTYPE:
            arguments.update(parse_qsl(self.body.decode('ascii')))
        return arguments

    @cached_property
    def cookies(self):
        cobj = SimpleCookie(self.cookie.decode('ascii', 'ignore'))
        return dict((k, cobj[k].value) for k in cobj)

    @classmethod
    def create(cls, resolver):
        return resolver.request


class WebException(Exception):
    """Base for all exceptions which render error code (and page) to client"""

    @abc.abstractmethod
    def default_response(self):
        pass

class Forbidden(WebException):

    def default_response(self):
        return (b'403 Forbidden',
                b'Content-Type\0text/html\0',
                b'<!DOCTYPE html>'
                b'<html>'
                    b'<head>'
                        b'<title>403 Forbidden</title>'
                    b'</head>'
                    b'<body>'
                    b'<h1>403 Forbidden</h1>'
                    b'</body>'
                b'</html>'
                )


class InternalError(WebException):

    def default_response(self):
        return (b'500 Internal Server Error',
                b'Content-Type\0text/html\0',
                b'<!DOCTYPE html>'
                b'<html>'
                    b'<head>'
                        b'<title>500 Internal Server Error</title>'
                    b'</head>'
                    b'<body>'
                    b'<h1>500 Internal Server Error</h1>'
                    b'</body>'
                b'</html>'
                )


class NotFound(WebException):

    def default_response(self):
        return (b'404 Not Found',
                b'Content-Type\0text/html\0',
                b'<!DOCTYPE html>'
                b'<html>'
                    b'<head>'
                        b'<title>404 Page Not Found</title>'
                    b'</head>'
                    b'<body>'
                    b'<h1>404 Page Not Found</h1>'
                    b'</body>'
                b'</html>'
                )


class MethodNotAllowed(WebException):

    def default_response(self):
        return (b'405 Method Not Allowed',
                b'Content-Type\0text/html\0',
                b'<!DOCTYPE html>'
                b'<html>'
                    b'<head>'
                        b'<title>405 Method Not Allowed</title>'
                    b'</head>'
                    b'<body>'
                    b'<h1>405 Method Not Allowed</h1>'
                    b'</body>'
                b'</html>'
                )


class Redirect(WebException):

    def __init__(self, location, status_code, status_text):
        self.statusline = '{:d} {}'.format(status_code, status_text)
        self.location = location

    def location_header(self):
        return b'Location\0' + self.location.encode('ascii') + b'\0'

    def headers(self):
        return (b'Content-Type\0text/html\0'
                + self.location_header())

    def default_response(self):
        return (self.statusline, self.headers(),
                '<!DOCTYPE html>'
                '<html>'
                    '<head>'
                        '<title>{0.statusline}</title>'
                    '</head>'
                    '<body>'
                    '<h1>{0.statusline}</h1>'
                    '<a href="{0.location}">Follow</a>'
                    '</body>'
                '</html>'.format(self)
                )


class CompletionRedirect(Redirect):
    """Temporary redirect which sends code 303

    With :param:`cookie` set it is often used for login forms. Without
    parameter set it is used to provide "success" page for various web forms
    and other non-idempotent actions
    """

    def __init__(self, location, cookie=None, *,
        status_code=303, status_text="See Other"):
        super().__init__(location,
            status_code=status_code, status_text=status_text)
        self.cookie = cookie

    def headers(self):
        sup = super().headers()
        if self.cookie is not None:
            sup += self.cookie.output(header='Set-Cookie\0',
                                      sep='\0').encode('ascii') + b'\0'
        return sup


class ChildNotFound(Exception):
    """Raised by resolve_local to notify that there is not such child"""


class NiceError(Exception):
    """Error that is safe to present to user"""


class BaseResolver(metaclass=abc.ABCMeta):

    _LEAF_METHODS = {_LEAF_METHOD}
    _RES_METHODS = {_RESOURCE_METHOD}
    resolver_class_attr = 'resolver_class'
    default_method = None

    def __init__(self, request, parent=None):
        self.request = request
        self.parent = parent
        self.resource = None

    @abc.abstractmethod
    def __next__(self):
        pass

    def __iter__(self):
        return self

    def child_fallback(self):
        raise NotFound()

    @abc.abstractmethod
    def set_args(self, args):
        pass

    def resolve(self, root):
        self.resource = node = root
        for name in self:
            # assert name is not None, "Wrong name from {!r}".format(self)
            try:
                node = node.resolve_local(name)
            except ChildNotFound:
                node = self.child_fallback()
            kind = getattr(node, '_zweb', None)
            if kind in self._LEAF_METHODS:
                return _dispatch_leaf(node, node.__self__, self)
            elif kind in self._RES_METHODS:
                newnode, tail = _dispatch_resource(node, node.__self__, self)
                if newnode is _INTERRUPT:
                    return tail  # tail is actual result in this case
                self.set_args(tail)
                self.resource = newnode
                res_class = getattr(newnode, self.resolver_class_attr, None)
                if res_class is None:
                    raise RuntimeError("Value {!r} returned from"
                        " {!r} is not a resource".format(newnode, node))
                node = newnode
                if not isinstance(self, res_class):
                    newres = res_class(self.request, self)
                    return newres.resolve(node)
            elif kind is _RESOURCE:
                self.resource = node
            else:
                log.debug("Wrong kind %r", kind)
                raise NotFound()  # probably impossible but ...


        if self.default_method is not None:
            meth = getattr(node, self.default_method, None)
            if(meth is not None
                and getattr(meth, '_zweb', None) in self._LEAF_METHODS):
                return _dispatch_leaf(node.index, node, self)

        raise NotFound()


class PathResolver(BaseResolver):

    _LEAF_METHODS = {_LEAF_METHOD, _LEAF_HTTP_METHOD}
    _RES_METHODS = {_RESOURCE_METHOD, _RES_HTTP_METHOD}
    resolver_class_attr = 'http_resolver_class'
    default_method = 'index'

    def __init__(self, request, parent=None):
        super().__init__(request, parent)
        path = request.parsed_uri.path.strip('/')
        if path:
            self.args = path.split('/')
        else:
            self.args = []
        self.kwargs = dict(request.form_arguments)

    def __next__(self):
        try:
            return self.args.pop(0)
        except IndexError:
            raise StopIteration()

    def set_args(self, args):
        self.args = list(args)


class MethodResolver(BaseResolver):

    _LEAF_METHODS = {_LEAF_METHOD, _LEAF_HTTP_METHOD}
    _RES_METHODS = {_RESOURCE_METHOD, _RES_HTTP_METHOD}
    resolver_class_attr = 'http_resolver_class'
    default_method = None

    def __init__(self, request, parent=None):
        super().__init__(request, parent)
        self.args = parent.args
        self.kwargs = dict(request.form_arguments)

    def __next__(self):
        return self.request.method.decode('ascii').upper()

    def child_fallback(self):
        raise MethodNotAllowed()

    def set_args(self, args):
        self.args = args


class WebsockResolver(BaseResolver):

    _LEAF_METHODS = {_LEAF_METHOD, _LEAF_WSOCK_METHOD}
    _RES_METHODS = {_RESOURCE_METHOD, _RES_WSOCK_METHOD}
    resolver_class_attr = 'websock_resolver_class'
    default_method = 'default'

    def __init__(self, request, parent=None):
        super().__init__(request, parent)
        self.parts = list(request.meth.split('.'))
        self.args = request.args
        self.kwargs = dict(request.kwargs)

    def __next__(self):
        try:
            return self.parts.pop(0)
        except IndexError:
            raise StopIteration()

    def set_args(self, args):
        self.args = args


class InternalRedirect(Exception, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def update_request(self, request):
        pass


class PathRewrite(InternalRedirect):

    def __init__(self, new_path):
        self.new_path = new_path.encode('ascii')

    def update_request(self, request):
        request.uri = self.new_path
        del request.parsed_uri


class ResourceInterface(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def resolve_local(self, name):
        """Returns child resource or method or raises ChildNotFound"""


@ResourceInterface.register
class Resource(object):
    http_resolver_class = PathResolver
    websock_resolver_class = WebsockResolver
    _zweb = _RESOURCE

    def resolve_local(self, name):
        if not name.isidentifier() or name.startswith('_'):
            raise ChildNotFound()
        target = getattr(self, name, None)
        if target is None:
            raise ChildNotFound()
        kind = getattr(target, '_zweb', None)
        if kind is not None:
            return target
        raise ChildNotFound()


@ResourceInterface.register
class DictResource(dict):

    http_resolver_class = PathResolver
    websock_resolver_class = WebsockResolver
    _zweb = _RESOURCE


    def resolve_local(self, name):
        try:
            return self[name]
        except KeyError:
            raise ChildNotFound()


class Site(object):

    def __init__(self, *, request_class, resources=()):
        self.request_class = request_class
        self.resources = resources

    def _resolve(self, request):
        for i in self.resources:
            res = i.http_resolver_class(request)
            try:
                return res.resolve(i)
            except NotFound as e:
                continue
        else:
            raise NotFound()

    def _safe_dispatch(self, request):
        while True:
            try:
                result = self._resolve(request)
            except InternalRedirect as e:
                e.update_request(request)
                continue
            except Exception as e:
                if not isinstance(e, WebException):
                    log.exception("Can't process request %r", request)
                    e = InternalError(e)
                try:
                    return self.error_page(e)
                except Exception:
                    log.exception("Can't make error page for %r", e)
                    return e.default_response()
            else:
                return result

    def error_page(self, e):
        return e.default_response()

    def __call__(self, *args):
        request = self.request_class(*args)
        result = self._safe_dispatch(request)
        if hasattr(result, 'zerogw_response'):
            return result.zerogw_response()
        else:
            return result


def _dispatch_resource(fun, self, resolver):
    preproc = getattr(fun, '_zweb_pre', ())
    result = None
    for prefun in preproc:
        result = prefun(self, resolver, *resolver.args, **resolver.kwargs)
        if result is not None:
            break
    if result is None:
        deco = getattr(fun, '_zweb_deco', None)
        if deco is not None:
            return deco(self, resolver,
                partial(fun._zweb_deco_callee, self, resolver),
                *resolver.args, **resolver.kwargs)
        else:
            try:
                args, tail, kw = fun._zweb_sig(resolver,
                    *resolver.args, **resolver.kwargs)
            except (TypeError, ValueError) as e:
                log.debug("Signature mismatch %r %r",
                    resolver.args, resolver.kwargs,
                    exc_info=e)  # debug
                raise NotFound()
            else:
                return fun(*args, **kw), tail
    else:
        for proc in fun._zweb_post:
            result = proc(self, resolver, result)
        return _INTERRUPT, result


class ReprHack(str):
    __slots__ = ()
    def __repr__(self):
        return str(self)


def _compile_signature(fun, partial):
    sig = inspect.signature(fun)
    fun_params = [
        inspect.Parameter('resolver',
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD)]
    args = []
    kwargs = []
    vars = {
        '__empty__': object(),
        }
    lines = []
    annotations = {}
    self = True
    varkw = False
    varpos = False

    for name, param in sig.parameters.items():
        ann = param.annotation
        if param.default is not inspect.Parameter.empty:
            vars[name + '_def'] = param.default
            defname = ReprHack('__empty__')
        else:
            defname = inspect.Parameter.empty
        if ann is not inspect.Parameter.empty:
            if isinstance(ann, type) and issubclass(ann, Sticker):
                lines.append('  {0} = {0}_create(resolver)'.format(name))
                vars[name + '_create'] = ann.create
            else:
                lines.append('  if {0} is __empty__:'.format(name))
                lines.append('    {0} = {0}_def'.format(name))
                lines.append('  else:')
                if isinstance(ann, type) and ann.__module__ == 'builtins':
                    lines.append('    {0} = {1}({0})'.format(
                        name, ann.__name__))
                    fun_params.append(param.replace(
                        default=defname))
                else:
                    lines.append('    {0} = {0}_type({0})'.format(name))
                    vars[name + '_type'] = ann
                    fun_params.append(param.replace(
                        annotation=ReprHack(name + '_type'),
                        default=defname))
        elif not self:
            fun_params.append(param.replace(default=defname))

        if param.kind == inspect.Parameter.VAR_KEYWORD:
            varkw = True
        elif param.kind == inspect.Parameter.VAR_POSITIONAL:
            varpos = name
        if param.kind == inspect.Parameter.KEYWORD_ONLY:
            kwargs.append('{0!r}: {0}'.format(name))
        elif not self:
            args.append(name)
        if self:
            self = False
    if not varpos and partial:
        for i, p in enumerate(fun_params):
            if p.kind == inspect.Parameter.KEYWORD_ONLY:
                fun_params.insert(i, inspect.Parameter('__tail__',
                    kind=inspect.Parameter.VAR_POSITIONAL))
                break
        else:
            fun_params.append(inspect.Parameter('__tail__',
                kind=inspect.Parameter.VAR_POSITIONAL))
    if not varkw:
        fun_params.append(inspect.Parameter('__kw__',
            kind=inspect.Parameter.VAR_KEYWORD))
    funsig = inspect.Signature(fun_params)
    lines.insert(0, 'def __sig__{}:'.format(funsig))
    if len(args) == 1:
        args = args[0] + ','
    else:
        args = ', '.join(args)
    if varpos:
        lines.append('  return ({}) + {}, (), {{{}}}'.format(
            args, varpos,  ', '.join(kwargs)))
    elif partial:
        lines.append('  return ({}), __tail__, {{{}}}'.format(
            args, ', '.join(kwargs)))
    else:
        lines.append('  return ({}), (), {{{}}}'.format(
            args, ', '.join(kwargs)))
    code = compile('\n'.join(lines), '__sig__', 'exec')
    exec(code, vars)
    return vars['__sig__']

def resource(fun):
    """Decorator to denote a method which returns resource to be traversed"""
    fun._zweb = _RESOURCE_METHOD
    fun._zweb_sig = _compile_signature(fun, partial=True)
    return fun


def http_resource(fun):
    """Decorator to denote a method which returns HTTP-only resource"""
    resource(fun)
    fun._zweb = _RES_HTTP_METHOD
    return fun


def websock_resource(fun):
    """Decorator to denote a method which returns Websocket-only resource"""
    resource(fun)
    fun._zweb = _RES_WSOCK_METHOD
    return fun


def _dispatch_leaf(fun, self, resolver):
    preproc = getattr(fun, '_zweb_pre', ())
    result = None
    for prefun in preproc:
        result = prefun(self, resolver, *resolver.args, **resolver.kwargs)
        if result is not None:
            break
    if result is None:
        deco = getattr(fun, '_zweb_deco', None)
        if deco is not None:
            result = deco(self, resolver,
                partial(fun._zweb_deco_callee, self, resolver),
                *resolver.args, **resolver.kwargs)
        else:
            try:
                args, tail, kw = fun._zweb_sig(resolver,
                    *resolver.args, **resolver.kwargs)
            except (TypeError, ValueError) as e:
                log.debug("Signature mismatch %r %r",
                    resolver.args, resolver.kwargs,
                    exc_info=e)  # debug
                raise NotFound()
            else:
                result = fun(*args, **kw)
    for proc in fun._zweb_post:
        result = proc(self, resolver, result)
    return result


def endpoint(fun):
    """Decorator to denote a method which returns some result to the user"""
    if not hasattr(fun, '_zweb_post'):
        fun._zweb_post = []
    fun._zweb = _LEAF_METHOD
    fun._zweb_sig = _compile_signature(fun, partial=False)
    return fun


def page(fun):
    """Decorator to denote a method which works only for http"""
    endpoint(fun)
    fun._zweb = _LEAF_HTTP_METHOD
    return fun


def method(fun):
    """Decorator to denote a method which works only for websockets"""
    endpoint(fun)
    fun._zweb = _LEAF_WSOCK_METHOD
    return fun


def postprocessor(fun):
    """A decorator that accepts method's output and processes it

    Works only on leaf nodes. Typical use cases are:

    * turn dict of variables into a JSON
    * render a template from the dict.
    """
    if not hasattr(fun, '_zweb_post'):
        fun._zweb_post = []
    def wrapper(proc):
        fun._zweb_post.append(proc)
        return fun
    return wrapper


def preprocessor(fun):
    """A decorators that runs before request

    When preprocessor returns None, request is processed as always. If
    preprocessor return not None it's return value treated as return value
    of a leaf node (even if it's resource) including executing all
    postprocessors.

    Works both on resources and leaf nodes. Typical use casees are:

    * access checking
    * caching
    """
    if not hasattr(fun, '_zweb_pre'):
        fun._zweb_pre = []
    def wrapper(proc):
        fun._zweb_pre.append(proc)
        return fun
    return wrapper


def decorator(fun):
    def wrapper(parser):
        olddec = getattr(fun, '_zweb_deco', None)
        oldcallee = getattr(fun, '_zweb_deco_callee', None)
        if olddec is None:
            def callee(self, resolver, *args, **kw):
                try:
                    args, tail, kw = fun._zweb_sig(resolver, *args, **kw)
                except (TypeError, ValueError) as e:
                    log.debug("Signature mismatch %r %r", args, kw,
                        exc_info=e)  # debug
                    raise NotFound()
                else:
                    return fun(self, *args, **kw)
        else:
            def callee(self, resolver, *args, **kw):
                return olddec(self, resolver,
                    partial(oldcallee, self, resolver),
                    *args, **kw)

        fun._zweb_deco = parser
        fun._zweb_deco_callee = callee
        return fun
    return wrapper


class Sticker(metaclass=abc.ABCMeta):
    """
    An object which is automatically put into arguments in the view if
    specified in annotation
    """
    __superseded = {}

    @classmethod
    @abc.abstractmethod
    def create(cls, resolver):
        """Creates an object of this class based on resolver"""

    @classmethod
    def supersede(cls, sub):
        oldsup = cls.__superseeded.get(cls, None)
        if oldsup is not None:
            if issubclass(sub, oldsup):
                pass  # just supersede it again
            elif issubclass(oldsup, sub):
                return  # already superseeded by more specific subclass
            else:
                raise RuntimeError("{!r} is already superseeded by {!r}"
                    .format(cls, oldsup))

        super().register(sub)
        cls.__superseded[cls] = sub


class WebsockCall(object):
    MESSAGE = 'message'
    CONNECT = 'connect'
    DISCONNECT = 'disconnect'
    HEARTBEAT = 'heartbeat'
    SYNC = 'sync'

    def __init__(self, msgs):
        cid = msgs[0]
        kind = msgs[1]
        meth = getattr(self, '_init_' + kind.decode('ascii'), None)
        if meth:
            meth(cid, *msgs[2:])

    def _init_message(self, cid, body):
        self.cid = cid
        self.args = json.loads(body.decode('utf-8'))
        self.meth = self.args.pop(0)
        self.kwargs = self.args.pop(0)
        self.request_id = self.kwargs.pop('#', None)
        self.kind = self.MESSAGE

    def _init_msgfrom(self, cid, cookie, body):
        self.cid = cid
        self.marker = cookie
        self.args = json.loads(body.decode('utf-8'))
        self.meth = self.args.pop(0)
        self.kwargs = self.args.pop(0)
        self.request_id = self.kwargs.pop('#', None)
        self.kind = self.MESSAGE

    def _init_connect(self, cid):
        self.cid = cid
        self.kind = self.CONNECT

    def _init_disconnect(self, cid, cookie=None):
        self.cid = cid
        self.marker = cookie
        self.kind = self.DISCONNECT

    def _init_heartbeat(self, server_id):
        self.server_id = server_id
        self.kind = self.HEARTBEAT

    def _init_sync(self, server_id, *users):
        self.server_id = server_id
        it = iter(users)
        self.users = dict(zip(it, it))
        self.kind = self.SYNC

    @classmethod
    def create(cls, resolver):
        return resolver.request


class Websockets(object):

    def __init__(self, *, resources=(), output=None):
        self.resources = resources
        self.output = output

    def _resolve(self, request):
        for i in self.resources:
            res = i.websock_resolver_class(request)
            try:
                return res.resolve(i)
            except NotFound as e:
                continue
        raise NotFound()

    def _safe_dispatch(self, request):
        try:
            result = self._resolve(request)
        except NiceError as e:
            return ['_error', e.client_data()]
        except NotFound as e:
            return ['_not_found', request.meth]
        except Exception as e:
            log.exception("Can't process request %r", request)
            return ['_internal_error']
        else:
            return ['_result', result]

    def __call__(self, *args):
        call = WebsockCall(args)
        if call.kind is WebsockCall.MESSAGE:
            result = self._safe_dispatch(call)
            if call.request_id is not None:
                result.insert(1, call.request_id)
                self.output.send(call, json.dumps(result))
        else:
            meth = getattr(self, 'handle_' + call.kind, None)
            if meth is not None:
                meth(call)


Sticker.register(Request)
Sticker.register(WebsockCall)



import socket
import random
import struct
import os
from collections import namedtuple
import logging
import abc
import time

from .core import gethub, Future, TimeoutError
from . import sleep


log = logging.getLogger(__name__)

type_to_code = {
    'A': b'\x00\x01',
    'NS': b'\x00\x02',
    'CNAME': b'\x00\x05',
    'PTR': b'\x00\x0C',
    'MX': b'\x00\x0F',
    'TXT': b'\x00\x10',
    'SRV': b'\x00\x21',
    }
IN_CLASS = b'\x00\x01'
code_to_type = {v: k for k, v in type_to_code.items()}
Header = namedtuple('Header', 'id flags qdcount ancount nscount arcount')
F_QR = 1 << 15
F_AA = 1 << 10
F_TC = 1 << 9
F_RD = 1 << 8
F_RA = 1 << 7
F_RCODE = 15
POINTER_MASK = 0b11000000


class DNSError(Exception):
    pass


class DNSFormatError(DNSError):
    code = 1


class DNSServerFailure(DNSError):
    code = 2


class DNSNameError(DNSError):
    code = 3


class DNSNotImplemented(DNSError):
    code = 4


class DNSRefused(DNSError):
    code = 5


code_to_error = {
    DNSFormatError.code: DNSFormatError,
    DNSServerFailure.code: DNSServerFailure,
    DNSNameError.code: DNSNameError,
    DNSNotImplemented.code: DNSNotImplemented,
    DNSRefused.code: DNSRefused,
    }


class Record(object):
    __slots__ = ('name', 'ttl')
    type = None
    klass = 'IN'

    def __init__(self, name, ttl):
        self.name = name
        self.ttl = ttl

    @abc.abstractmethod
    def parse_from(self, data, pos, ln):
        pass

    def __repr__(self):
        return '<RR {} {}>'.format(self.type,
            ', '.join(k + '=' + repr(getattr(self, k))
            for k in self.__slots__))


class ARecord(Record):
    __slots__ = ('ip',)
    type = 'A'

    def parse_from(self, data, pos, ln):
        self.ip = socket.inet_ntoa(data[pos:pos+ln])


class MXRecord(Record):
    __slots__ = ('priority', 'server')
    type = 'MX'

    def parse_from(self, data, pos, ln):
        self.priority, = struct.unpack_from('>H', data, pos)
        self.server, _ = _read_name(data, pos+2)


class CNAMERecord(Record):
    __slots__ = ('canonical_name',)
    type = 'CNAME'

    def parse_from(self, data, pos, ln):
        self.canonical_name = _read_name(data, pos)[0]


class NSRecord(Record):
    __slots__ = ('authoritative_name',)
    type = 'NS'

    def parse_from(self, data, pos, ln):
        self.authoritative_name = _read_name(data, pos)[0]


class PTRRecord(Record):
    __slots__ = ('pointer_target',)
    type = 'PTR'

    def parse_from(self, data, pos, ln):
        self.pointer_target = _read_name(data, pos)[0]


class TXTRecord(Record):
    __slots__ = ('text',)
    type = 'TXT'

    def parse_from(self, data, pos, ln):
        self.text = data[pos:pos+ln].decode('utf-8')


class SRVRecord(Record):
    __slots__ = ('priority', 'weight', 'port', 'target')
    type = 'SRV'

    def parse_from(self, data, pos, ln):
        fields = struct.unpack_from('>HHH', data, pos)
        self.priority, self.weight, self.port = fields
        self.target, _ = _read_name(data, pos+6)


rr_types = {
    'A': ARecord,
    'CNAME': CNAMERecord,
    'NS': NSRecord,
    'PTR': PTRRecord,
    'TXT': TXTRecord,
    'MX': MXRecord,
    'SRV': SRVRecord,
    }


def _read_name(data, pos):
    name = []
    while True:
        ln = data[pos]
        if ln & POINTER_MASK == POINTER_MASK:
            off = ((ln & ~POINTER_MASK) << 8) | data[pos+1]
            name.append(_read_name(data, off)[0])
            return '.'.join(name), pos+2
        else:
            pos += 1
            if not ln:
                return '.'.join(name), pos
            part = data[pos:pos+ln].decode('ascii')
        name.append(part)
        pos += ln


class Resolver(object):

    def __init__(self, config, max_ttl=86400):
        self.config = config
        self.max_ttl = max_ttl
        self.cur_requests = {}
        self.udp_requests = {}
        self.cache = {}
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sock.setblocking(0)
        hub = gethub()
        hub.do_spawnhelper(self._receiver)

    def request(self, name, typ):
        key = name, typ
        fut = self.cur_requests.get(key, None)
        if fut is not None:
            return fut
        fut = Future()
        self.cur_requests[key] = fut
        try:
            self._do_resolve(name, typ, fut)
        finally:
            del self.cur_requests[key]
        return fut

    def resolve(self, name, typ='A'):
        if self.config.check():
            self.cache.clear()
        cache = self.config.hosts.get(name, None)
        if cache is not None:
            return [cache]
        cache = self.cache.get((name, typ))
        if cache is not None:
            return cache
        return self.request(name, typ).get()

    def _do_resolve(self, name, typ, fut):
        query = bytearray()
        query.extend(b'\x00\x00'  # dummy request id for now
                     b'\x01\x00'  # flags: recursion only
                     b'\x00\x01'  # only single query
                     b'\x00\x00'  # no answers
                     b'\x00\x00'  # no authority
                     b'\x00\x00') # no additional
        for part in name.split('.'):
            part = part.encode('ascii')
            query.append(len(part))
            query.extend(part)
        query.append(0)
        query.extend(type_to_code[typ])
        query.extend(IN_CLASS)
        ns_to_try = self.config.ns[:]
        for n in ns_to_try:
            rid = random.randrange(65536)
            struct.pack_into('>H', query, 0, rid)
            key = rid, name, typ
            self.udp_requests[key] = fut
            try:
                try:
                    self._sock.sendto(query, (n, 53))
                except OSError as e:
                    continue
                try:
                    fut.get(timeout=self.config.options['timeout'])
                except TimeoutError:
                    continue
                else:
                    break
            finally:
                del self.udp_requests[key]
        if fut.check():
            fut.throw(DNSRefused("No server available"))

    def _receiver(self):
        while True:
            gethub().do_read(self._sock)
            try:
                data, (host, port) = self._sock.recvfrom(512)
            except OSError:
                continue
            if port != 53 or host not in self.config.ns:
                continue  # some garbage received
            try:
                packet = self._parse_packet(data)
            except Exception as e:
                log.exception('Error parsing DNS packet', exc_info=e)
                continue
            if packet is None:
                continue

    def _parse_packet(self, data):
        head = Header(*struct.unpack_from('>6H', data))
        if not (head.flags & F_QR):
            return  # got question ?
        if head.qdcount != 1:
            return  # we don't ask several questions at once
        name, pos = _read_name(data, 12)
        typ = code_to_type.get(data[pos:pos+2], None)
        if typ is None:
            return  # unsupported type
        if data[pos+2:pos+4] != IN_CLASS:
            return  # we only request IN class
        pos += 4
        fut = self.udp_requests.get((head.id, name, typ), None)
        if fut is None or not fut.check():
            return  # late or unsolicited reply
        err = head.flags & F_RCODE
        if err:
            exc = code_to_error.get(err)
            if exc is not None:
                fut.throw(exc())
            else:
                fut.throw(DNSError("Resolving error: {}".format(err)))
            return
        rrs = []
        for i in range(head.ancount):
            rr, pos = self._read_rr(data, pos)
            rrs.append(rr)
        fut.set(rrs)

    @staticmethod
    def _read_rr(data, pos):
        name, pos = _read_name(data, pos)
        typ = code_to_type.get(data[pos:pos+2])
        if typ is None:
            raise ValueError("Wrong RR type {!r}".format(data[pos:pos+2]))
        cls = data[pos+2:pos+4]
        assert cls == IN_CLASS
        pos += 4
        ttl, ln = struct.unpack_from('>LH', data, pos)
        pos += 6
        cls = rr_types[typ]
        rr = cls(name, ttl)
        rr.parse_from(data, pos, ln)
        return rr, pos+ln

    # CONVENTIONAL FUNCTIONS

    def gethostbyname(self, name):
        records = self.resolve(name)
        by_name = {}
        for rec in records:
            if rec.name == name and isinstance(rec, ARecord):
                return rec.ip
            by_name[rec.name] = rec
        while True:
            rec = by_name.get(name, None)
            if isinstance(rec, CNAMERecord):
                name = rec.canonical_name
            elif isinstance(rec, ARecord):
                return rec.ip
            else:
                raise DNSServerFailure(records)


class Config(object):

    def __init__(self, check_files_time=1):
        self.check_files_time = check_files_time
        self.files_to_check = {}
        self.clear()

    def clear(self):
        self.hosts = {}
        self.ns = []
        self.search = []
        # self.sortlist = []  # TODO(tailhook) too rare to implement now
        self.options = {
            'ndots': 1,
            'timeout': 1,  # what is a good timeout?
            'attempts': 2,
            'rotate': False,
            }

    def check(self):
        if self.last_check + self.check_files_time < time.time():
            for fn, mtime in self.files_to_check.items():
                if os.path.getmtime(fn) != mtime:
                    break
            else:
                return False
            self.clear()
            self.load_from_files(
                hosts=self.hosts_file,
                resolv=self.resolv_file)
            return True
        return False

    @classmethod
    def system_config(cls):
        self = cls()
        self.load_from_files()
        return self

    def load_from_files(self, hosts='/etc/hosts', resolv='/etc/resolv.conf'):
        self.last_check = time.time()
        self.hosts_file = hosts
        self.resolv_file = resolv
        self.read_hosts_file(hosts)
        self.read_resolv_file(resolv)
        self.apply_environment()

    def read_hosts_file(self, fn='/etc/hosts'):
        with open(fn, 'rt') as f:
            self.files_to_check[fn] = os.path.getmtime(fn)
            for line in f:
                line = line.strip()
                if not line or line.startswith((';', '#')):
                    continue
                ip, *hosts = line.split()
                for h in hosts:
                    self.hosts[h] = ip

    def read_resolv_file(self, fn='/etc/resolv.conf'):
        with open(fn, 'rt') as f:
            self.files_to_check[fn] = os.path.getmtime(fn)
            for line in f:
                line = line.strip()
                if line.startswith((';', '#')):
                    continue
                option, arg = line.split(None, 1)
                if option == 'nameserver':
                    self.ns.append(arg)
                elif option == 'domain':
                    self.search = [arg]
                elif option == 'search':
                    self.search.append(arg)
                elif option == 'options':
                    self.parse_options(arg)
                else:
                    warnings.warn('Wrong option {!r} in {!r}'.format(
                        option, fn))

    def apply_environment(self, env=os.environ):
        if 'LOCALDOMAIN' in env:
            self.search = env['LOCALDOMAIN'].split()
        if 'RES_OPTIONS' in env:
            self.options.update(self.parse_options(env['RES_OPTIONS']))

    def parse_options(self, optstr):
        for word in optstr.split():
            if ':' in word:
                key, val = word.split(':', 1)
                self.options[key] = int(val)
            else:
                self.options[word] = True




import socket
import errno
import struct
import os.path
import hashlib
import warnings
import string
from datetime import date, time, datetime, timedelta
from collections import namedtuple
from decimal import Decimal

from .core import gethub, Lock, Future
from . import channel
from .util import marker_object, setcloexec


PREPARED_STMT = marker_object("PREPARED_STMT")
PREPARED_PARAMS = marker_object("PREPARED_PARAMS")
PREPARED_COLS = marker_object("PREPARED_COLS")
QUERY = marker_object("QUERY")
QUERY_FIELDS = marker_object("QUERY_FIELDS")
QUERY_ROWDATA = marker_object("QUERY_ROWDATA")
HANDSHAKE = marker_object("HANDSHAKE")
FLAG_BINARY = 0x0080
FLAG_UNSIGNED = 0x0020

OK_PACKET = (bytearray(b'\x00\x00\x00\x02\x00\x00\x00'),)
FIELD_STR = struct.Struct('<HLBHB')

RESERVED = frozenset((
    'accessible', 'add', 'all', 'alter', 'analyze', 'and', 'as', 'asc',
    'asensitive', 'before', 'between', 'bigint', 'binary', 'blob', 'both',
    'by', 'call', 'cascade', 'case', 'change', 'char', 'character', 'check',
    'collate', 'column', 'condition', 'constraint', 'continue', 'convert',
    'create', 'cross', 'current_date', 'current_time', 'current_timestamp',
    'current_user', 'cursor', 'database', 'databases', 'day_hour',
    'day_microsecond', 'day_minute', 'day_second', 'dec', 'decimal', 'declare',
    'default', 'delayed', 'delete', 'desc', 'describe', 'deterministic',
    'distinct', 'distinctrow', 'div', 'double', 'drop', 'dual', 'each', 'else',
    'elseif', 'enclosed', 'escaped', 'exists', 'exit', 'explain', 'false',
    'fetch', 'float', 'float4', 'float8', 'for', 'force', 'foreign', 'from',
    'fulltext', 'general', 'grant', 'group', 'having', 'high_priority',
    'hour_microsecond', 'hour_minute', 'hour_second', 'if', 'ignore',
    'ignore_server_ids', 'in', 'index', 'infile', 'inner', 'inout',
    'insensitive', 'insert', 'int', 'int1', 'int2', 'int3', 'int4', 'int8',
    'integer', 'interval', 'into', 'is', 'iterate', 'join', 'key', 'keys',
    'kill', 'leading', 'leave', 'left', 'like', 'limit', 'linear', 'lines',
    'load', 'localtime', 'localtimestamp', 'lock', 'long', 'longblob',
    'longtext', 'loop', 'low_priority', 'master_heartbeat_period',
    'master_ssl_verify_server_cert', 'match', 'maxvalue', 'mediumblob',
    'mediumint', 'mediumtext', 'middleint', 'minute_microsecond',
    'minute_second', 'mod', 'modifies', 'natural', 'not', 'no_write_to_binlog',
    'null', 'numeric', 'on', 'optimize', 'option', 'optionally', 'or', 'order',
    'out', 'outer', 'outfile', 'precision', 'primary', 'procedure', 'purge',
    'range', 'read', 'reads', 'read_write', 'real', 'references', 'regexp',
    'release', 'rename', 'repeat', 'replace', 'require', 'resignal',
    'restrict', 'return', 'revoke', 'right', 'rlike', 'schema', 'schemas',
    'second_microsecond', 'select', 'sensitive', 'separator', 'set', 'show',
    'signal', 'slow[d]', 'smallint', 'spatial', 'specific', 'sql',
    'sqlexception', 'sqlstate', 'sqlwarning', 'sql_big_result',
    'sql_calc_found_rows', 'sql_small_result', 'ssl', 'starting',
    'straight_join', 'table', 'terminated', 'then', 'tinyblob', 'tinyint',
    'tinytext', 'to', 'trailing', 'trigger', 'true', 'undo', 'union', 'unique',
    'unlock', 'unsigned', 'update', 'usage', 'use', 'using', 'utc_date',
    'utc_time', 'utc_timestamp', 'values', 'varbinary', 'varchar',
    'varcharacter', 'varying', 'when', 'where', 'while', 'with', 'write',
    'xor', 'year_month', 'zerofill'
    ))


def _read_lcb(buf, pos=0):
    num = buf[pos]
    if num < 251:
        return num, pos+1
    elif num == 251:
        return None, pos+1
    elif num == 252:
        return struct.unpack_from('<H', buf, pos+1)[0], pos+3
    elif num == 253:
        return buf[pos+1] + (buf[pos+2] << 8) + (buf[pos+3] << 16), pos+4
    elif num == 254:
        return struct.unpack_from('<Q', buf, pos+1)[0], pos+9


def _read_lcbytes(buf, pos=0):
    num = buf[pos]
    pos += 1
    if num < 251:
        return buf[pos:pos+num], pos+num
    elif num == 251:
        return None, pos
    elif num == 252:
        num = struct.unpack_from('<H', buf, pos)[0]
        pos += 2
    elif num == 253:
        num = buf[pos] + (buf[pos+1] << 8) + (buf[pos+2] << 16)
        pos += 3
    elif num == 254:
        num = struct.unpack_from('<Q', buf, pos)[0]
        pos += 8
    return buf[pos:pos+num], pos+num


def _read_lcstr(buf, pos=0):
    num = buf[pos]
    pos += 1
    if num < 251:
        pass
    elif num == 251:
        return None, pos
    elif num == 252:
        num = struct.unpack_from('<H', buf, pos)[0]
        pos += 2
    elif num == 253:
        num = buf[pos] + (buf[pos+1] << 8) + (buf[pos+2] << 16)
        pos += 3
    elif num == 254:
        num = struct.unpack_from('<Q', buf, pos)[0]
        pos += 8
    return buf[pos:pos+num].decode('utf-8'), pos+num


def _write_lcbytes(buf, data):
    ln = len(data)
    if ln <= 250:
        buf.append(ln)
    elif ln <= 0xFFFF:
        buf += b'\xfc' + struct.pack('<H', ln)
    elif ln <= 0xFFFFFF:
        buf.append(b'\xfd')
        buf.append(ln & 0xFF)
        buf.append((ln >> 8) & 0xFF)
        buf.append(ln >> 16)
    else:
        buf.append(b'\xfe')
        buf += struct.pack('<Q', ln)
    buf += data


def _read_bintime(buf, pos):
    ln = buf[pos]
    assert ln >= 8
    sign, days, hour, min, sec = struct.unpack_from('<BLBBB', buf, pos+1)
    assert not sign and not days, 'Timedeltas are not supported'
    return time(hour, min, sec), pos+ln+1


def _read_bindate(buf, pos):
    ln = buf[pos]
    assert ln >= 4
    year, month, day = struct.unpack_from('<H2B', buf, pos+1)
    return date(year, month, day), pos+ln+1


def _read_bindatetime(buf, pos):
    ln = buf[pos]
    assert ln >= 4
    if ln > 7:
        assert ln == 11
        year, month, day, hour, min, sec, ms \
            = struct.unpack_from('<H5BL',buf,pos+1)
    elif ln > 4:
        ms = 0
        year, month, day, hour, min, sec = struct.unpack_from('<H5B',buf,pos+1)
    else:
        ms = 0
        hour = 0
        min = 0
        sec = 0
        year, month, day = struct.unpack_from('<H2B', buf, pos+1)
    return datetime(year, month, day, hour, min, sec, ms), pos+ln+1


FIELD_MAPPING = {
    0x00: lambda a: Decimal(str(a, 'utf-8')),
    0x01: int,
    0x02: int,
    0x03: int,
    0x04: float,
    0x05: float,
    0x06: lambda a: None,
    0x08: int,
    0x09: int,
    0x0a: lambda a: datetime.strptime(str(a, 'ascii'), '%Y-%m-%d').date(),
    0x0b: lambda a: datetime.strptime(str(a, 'ascii'), '%H:%M:%S').time(),
    0x0c: lambda a: datetime.strptime(str(a, 'ascii'), '%Y-%m-%d %H:%M:%S'),
    0x0d: int,
    0x0f: lambda a: str(a, 'utf-8'),
    0xf6: lambda a: str(a, 'utf-8'),
    0xf7: lambda a: str(a, 'utf-8'),
    0xf8: lambda a: set(s.decode('utf-8').split(',')),
    0xf9: bytes,
    0xfa: bytes,
    0xfb: bytes,
    0xfc: bytes,
    0xfd: lambda a: str(a, 'utf-8'),
    0xfe: lambda a: str(a, 'utf-8'),
    }
FIELD_BIN_READERS = { # type, unsigned property
    (0x01, False): lambda buf, pos: (struct.unpack_from('<b', buf, pos)[0], pos+1),
    (0x02, False): lambda buf, pos: (struct.unpack_from('<h', buf, pos)[0], pos+2),
    (0x03, False): lambda buf, pos: (struct.unpack_from('<l', buf, pos)[0], pos+4),
    (0x04, False): lambda buf, pos: (struct.unpack_from('<f', buf, pos)[0], pos+4),
    (0x05, False): lambda buf, pos: (struct.unpack_from('<d', buf, pos)[0], pos+8),
    (0x08, False): lambda buf, pos: (struct.unpack_from('<q', buf, pos)[0], pos+8),
    (0x01, True): lambda buf, pos: (buf[pos], pos+1),
    (0x02, True): lambda buf, pos: (struct.unpack_from('<H', buf, pos)[0], pos+2),
    (0x03, True): lambda buf, pos: (struct.unpack_from('<L', buf, pos)[0], pos+4),
    (0x08, True): lambda buf, pos: (struct.unpack_from('<Q', buf, pos)[0], pos+8),
    (0x0a, False): _read_bindate,
    (0x0b, False): _read_bintime,
    (0x0c, False): _read_bindatetime,
    (0x0f, False): _read_lcstr,
    (0xf6, False): _read_lcstr,  # new decimal
    (0xf7, False): _read_lcstr,
    (0xf9, False): _read_lcbytes,
    (0xfa, False): _read_lcbytes,
    (0xfb, False): _read_lcbytes,
    (0xfc, False): _read_lcbytes,
    (0xfd, False): _read_lcstr,
    (0xfe, False): _read_lcstr,
    }


class MysqlError(Exception):

    def __init__(self, errno, sqlstate, message):
        self.errno = errno
        self.sqlstate = sqlstate
        self.message = message

    def __str__(self):
        return '({}:{}) {}'.format(self.errno, self.sqlstate, self.message)


class ColumnName(str):
    __slots__ = ()


def str_format(s):
    return ("'%s'" % s
        .replace('\\', r'\\')
        .replace('\0', r'\0')
        .replace('\'', r"\'")
        .replace('\"', r'\"')
        .replace('\b', r'\b')
        .replace('\n', r'\n')
        .replace('\r', r'\r')
        .replace('\t', r'\t')
        .replace('\x1A', r'\Z')
        )


class Formatter(string.Formatter):

    def format_field(self, value, format_spec):
        if isinstance(value, ColumnName):
            if format_spec:
                raise ValueError("Unknown format specification {0!r}"
                    .format(format_spec))
            if value in RESERVED:
                return '`{}`'.format(value.replace('`', '``'))
            return value
        elif format_spec:
            return str_format(format(value, format_spec))
        else:
            if isinstance(value, (int, float)):
                return str(value)
            elif isinstance(value, date):
                if isinstance(value, datetime):
                    return format(value, "'%Y-%m-%d %H:%M:%S'")
                else:
                    return format(value, "'%Y-%m-%d'")
            elif isinstance(value, timedelta):
                return ("INTERVAL '{0} {1:02d}:{2:02d}:{3:02d}.{4:06d}'"
                    " DAY_MICROSECOND".format(value.days,
                    value.seconds // 3600, value.seconds // 60 % 60,
                    value.seconds % 60, value.microseconds))
            elif isinstance(value, time):
                return format(value, "'%H:%M:%S'")
            elif isinstance(value, str):
                return str_format(value)
            elif value is None:
                return 'NULL'
            elif isinstance(value, bytes):
                raise ValueError("Use prepared statements to work with bytes")
            else:
                return str_format(str(value))

    def convert_field(self, value, conversion):
        if conversion in {'c', 'f', 't'}:
            # column, field, table -- same quoting
            return ColumnName(value)
        else:
            return super().convert_field(value, conversion)


_formatter = Formatter()
mysql_format = _formatter.format
mysql_vformat = _formatter.vformat

_Field = namedtuple('_Field', 'catalog db table org_table name org_name'
        ' charsetnr length type flags decimals default bin_key')

class Field(_Field):
    __slots__ = ()

    @classmethod
    def parse_packet(cls, packet, pos=0):
        catalog, pos = _read_lcstr(packet, pos)
        db, pos = _read_lcstr(packet, pos)
        table, pos = _read_lcstr(packet, pos)
        org_table, pos = _read_lcstr(packet, pos)
        name, pos = _read_lcstr(packet, pos)
        org_name, pos = _read_lcstr(packet, pos)
        pos += 1
        charset, length, type, flags, decimals \
            = FIELD_STR.unpack_from(packet, pos)
        pos += FIELD_STR.size + 2
        if len(packet) > pos:
            default = _read_lcstr(packet, pos)
        else:
            default = None
        return cls(catalog, db, table, org_table, name, org_name,
            charset, length, type, flags, decimals, default,
            (type, bool(flags & FLAG_UNSIGNED)))


class Resultset(object):

    def __init__(self, reply, nfields, extra):
        self.nfields = nfields
        self.extra = extra
        self.fields = [Field.parse_packet(fp) for fp in reply[1:nfields+1]]
        self.reply = reply

    def __iter__(self):
        raise RuntimeError("Can directly iterate only on resultset from"
            " prepared statement. Use dicts() or tuples() methods")

    def _parse_row(self, rpacket):
        pos = 0
        for f in self.fields:
            col, pos = _read_lcbytes(rpacket, pos)
            if col is None:
                yield None
            else:
                cvt = FIELD_MAPPING.get(f.type)
                if cvt is None:
                    raise RuntimeError('{} is not supported'
                        .format(f.type))
                yield cvt(col)

    def dicts(self):
        for rpacket in self.reply[self.nfields+2:-1]:
            yield {f.name: val for f, val in zip(self.fields,
                                                 self._parse_row(rpacket))}

    def tuples(self):
        for rpacket in self.reply[self.nfields+2:-1]:
            yield tuple(self._parse_row(rpacket))


class BinaryResultset(Resultset):

    def __init__(self, cls, reply, nfields, extra):
        super().__init__(reply, nfields, extra)
        self.row_class = cls
        self.fields = [Field.parse_packet(fp) for fp in reply[1:nfields+1]]

    def __iter__(self):
        for rpacket in self.reply[self.nfields+2:-1]:
            yield self.row_class(*self._parse_row(rpacket))

    def _parse_row(self, rpacket):
        nbytes = (self.nfields+2+7)//8
        row = []
        pos = 1+nbytes
        mask = rpacket[1:pos]
        for i, f in enumerate(self.fields, 2):
            if mask[i//8] & (1 << (i % 8)):
                yield None
            else:
                read = FIELD_BIN_READERS.get(f.bin_key)
                if read is None:
                    raise RuntimeError('{} is not supported'.format(f.type))
                val, pos = read(rpacket, pos)
                yield val


class PreparedStatement(object):

    def __init__(self, text, id, fields, params):
        self.text = text
        self.id = id
        self.bound = False
        self.fields = fields
        self.params = params
        if self.fields:
            self.row_class = namedtuple('Row', [f.name for f in fields])

    def write_binding(self, buf):
        buf += b'\x01'
        for param in self.params:
            buf += struct.pack('<H', param.type)
        return buf


execute_result = namedtuple('ExecuteResult', 'insert_id affected_rows')


class Capabilities(object):

    def __init__(self, num):
        self.long_password = bool(num & 1)
        self.found_rows = bool(num & 2)
        self.long_flag = bool(num & 4)
        self.connect_with_db = bool(num & 8)
        self.no_schema = bool(num & 16)
        self.compress = bool(num & 32)
        self.odbc = bool(num & 64)
        self.local_files = bool(num & 128)
        self.ignore_space = bool(num & 256)
        self.protocol_41 = bool(num & 512)
        self.interactive = bool(num & 1024)
        self.ssl = bool(num & 2048)
        self.ignore_sigpipe = bool(num & 4096)
        self.transactions = bool(num & 8192)
        self.secure_connection = bool(num & 32768)
        self.multi_statements = bool(num & 65536)
        self.multi_results = bool(num & 131072)

    def to_int(self):
        num = 0
        if self.long_password: num |= 1
        if self.found_rows: num |= 2
        if self.long_flag: num |= 4
        if self.connect_with_db: num |= 8
        if self.no_schema: num |= 16
        if self.compress: num |= 32
        if self.odbc: num |= 64
        if self.local_files: num |= 128
        if self.ignore_space: num |= 256
        if self.protocol_41: num |= 512
        if self.interactive: num |= 1024
        if self.ssl: num |= 2048
        if self.ignore_sigpipe: num |= 4096
        if self.transactions: num |= 8192
        if self.secure_connection: num |= 32768
        if self.multi_statements: num |= 65536
        if self.multi_results: num |= 131072
        return num


def _parse_error(packet):
    if packet[3] == 35: # it's '#'
        errno, code = struct.unpack_from('<xHx5s', packet)
        raise MysqlError(errno,
                         code.decode('ascii'),
                         packet[9:].decode('utf-8'))
    else:
        errno, = struct.unpack_from('<xH', packet)
        raise MysqlError(errno, '?', packet[3:].decode('utf-8'))


class Channel(channel.PipelinedReqChannel):
    BUFSIZE = 16384

    def connect(self, host, port, unixsock, user, password, database):
        try:
            return self._connect(host, port, unixsock,
                                 user, password, database)
        except Exception:
            self._alive = False
            raise

    def _connect(self, host, port, unixsock, user, password, database):
        sock = None
        if host == 'localhost':
            if os.path.exists(unixsock):
                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                addr = unixsock
        if sock is None:
            sock = socket.socket(socket.AF_INET,
                socket.SOCK_STREAM, socket.IPPROTO_TCP)
            addr = (host, port)
        setcloexec(sock)
        self._sock = sock
        self._sock.setblocking(0)
        fut = Future()
        self._producing.append((HANDSHAKE, fut))
        try:
            self._sock.connect(addr)
        except socket.error as e:
            if e.errno == errno.EINPROGRESS:
                gethub().do_write(self._sock)
            else:
                raise
        self._start()
        handshake, = fut.get()

        assert handshake[0] == 10, "Wrong protocol version {}".format(
            handshake[0])
        prefix, suffix = handshake[0:].split(b'\0', 1)
        self.thread_id, scramble, caplow, self.language, \
        self.status, caphigh, scrlen = struct.unpack_from('<L8sxHBHHB', suffix)
        scramble += suffix[31:suffix.index(b'\x00', 31)]
        self.capabilities = Capabilities((caphigh << 16) + caplow)
        assert self.capabilities.protocol_41, "Old protocol is not supported"
        assert self.capabilities.connect_with_db
        self.capabilities.odbc = False
        self.capabilities.compress = False
        self.capabilities.multi_statement = False
        self.capabilities.multi_results = False
        self.capabilities.ssl = False
        self.capabilities.transactions = False
        self.capabilities.no_schema = False  # for "show tables" to work
        buf = bytearray(b'\x00\x00\x00\x01')
        buf += struct.pack('<L4sB23s',
            self.capabilities.to_int()&0xFFFF,
            b'\x8f\xff\xff\xff',
            33, # utf-8 character set with general collation
            b'\x00'*23)
        buf += user.encode('ascii')
        buf += b'\x00'
        if password:
            buf += b'\x14'
            hash1 = hashlib.sha1(password.encode('ascii')).digest()
            hash2 = hashlib.sha1(scramble
                + hashlib.sha1(hash1).digest()).digest()
            buf += bytes(a^b for a, b in zip(hash1, hash2))
        else:
            buf += b'\x00'
        buf += database.encode('ascii')
        buf += b'\x00'
        value = self.request(buf, HANDSHAKE).get()
        if value[-1][0] == 0xff:
            _parse_error(value[-1])
        assert value == OK_PACKET, value

    def sender(self):
        buf = bytearray()

        add_chunk = buf.extend
        wait_write = gethub().do_write
        sock = self._sock

        while True:
            if not buf:
                self.wait_requests()
            wait_write(sock)
            for chunk in self.get_pending_requests():
                ln = len(chunk)-4
                chunk[0] = ln & 0xFF
                chunk[1] = (ln >> 8) & 0xFF
                chunk[2] = (ln >> 16) & 0xFF
                add_chunk(chunk)
            try:
                bytes = sock.send(buf)
            except socket.error as e:
                if e.errno in (errno.EAGAIN, errno.EINTR):
                    continue
                elif e.errno in (errno.EPIPE, errno.ECONNRESET):
                    raise EOFError()
                else:
                    raise
            if not bytes:
                raise EOFError()
            del buf[:bytes]

    def receiver(self):
        buf = bytearray()

        sock = self._sock
        wait_read = gethub().do_read
        add_chunk = buf.extend
        pos = 0
        current = []

        while True:
            if pos*2 > len(buf):
                del buf[:pos]
                pos = 0
            wait_read(sock)
            try:
                bytes = sock.recv(self.BUFSIZE)
                if not bytes:
                    raise EOFError()
                add_chunk(bytes)
            except socket.error as e:
                if e.errno in (errno.EAGAIN, errno.EINTR):
                    continue
                elif e.errno in (errno.EPIPE, errno.ECONNRESET):
                    raise EOFError()
                else:
                    raise
            while len(buf)-pos >= 4:
                length = buf[pos] + (buf[pos+1] << 8) + (buf[pos+2] << 16)
                if len(buf)-pos < length+4:
                    break
                num = buf[pos+3]
                ptype = buf[pos+4]
                self.produce(buf[pos+4:pos+length+4])
                pos += length+4

    def produce(self, value):
        if not self._alive:
            raise channel.ShutdownException()
        state = self.__dict__.setdefault('_state', self._producing[0][0])
        if state is HANDSHAKE:
            del self._state
            self._producing.popleft()[1].set((value,))
        elif value[0] == 0xff:
            self._cur_producing.append(value)
            self._do_produce()
        elif state is QUERY:
            if value[0] == 0x00:
                del self._state
                self._producing.popleft()[1].set((value,))
            else:
                self._cur_producing.append(value)
                self._state = QUERY_FIELDS
        elif state is QUERY_FIELDS:
            self._cur_producing.append(value)
            if value[0] == 0xfe:
                self._state = QUERY_ROWDATA
        elif state is QUERY_ROWDATA:
            self._cur_producing.append(value)
            if value[0] == 0xfe:
                self._do_produce()
        elif state is PREPARED_STMT:
            self._cur_producing.append(value)
            cols, params = struct.unpack_from('<5xHH', value)
            if params:
                self._state = PREPARED_PARAMS
            elif cols:
                self._state = PREPARED_COLS
            else:
                self._do_produce()
        elif state is PREPARED_PARAMS:
            self._cur_producing.append(value)
            if value[0] == 0xfe:
                cols, = struct.unpack_from('<5xH', self._cur_producing[0])
                if cols:
                    self._state = PREPARED_COLS
                else:
                    self._do_produce()
        elif state is PREPARED_COLS:
            self._cur_producing.append(value)
            if value[0] == 0xfe:
                self._do_produce()
        else:
            raise NotImplementedError(state)

    def _do_produce(self):
        res = tuple(self._cur_producing)
        del self._cur_producing[:]
        del self._state
        self._producing.popleft()[1].set(res)


class Mysql(object):

    def __init__(self, host='localhost', port=3306,
                       unixsock='/var/run/mysqld/mysqld.sock',
                       user='root', password='', database='test'):
        self._channel = None
        self._channel_lock = Lock()
        self.host = host
        self.port = port
        self.unixsock = unixsock
        self.user = user
        self.password = password
        self.database = database

    def channel(self):
        if not self._channel:
            with self._channel_lock:
                if not self._channel:
                    self._prepared = {}  # empty on reconnect
                    chan = Channel()
                    chan.connect(self.host, self.port,
                        unixsock=self.unixsock,
                        user=self.user, password=self.password,
                        database=self.database)
                    self._channel = chan
        return self._channel

    def execute(self, query, *args, **kw):
        if args or kw:
            query = mysql_vformat(query, args, kw)
        chan = self.channel()
        buf = bytearray(b'\x00\x00\x00\x00')
        buf += b'\x03'
        buf += query.encode('utf-8')
        reply = chan.request(buf, QUERY).get()
        if reply[-1][0] == 0xff:
            _parse_error(reply[-1])
        return self._parse_execute(reply, query)

    def _parse_execute(self, reply, query):
        assert len(reply) == 1, "Use query for queries that return result set"
        reply = reply[0]
        assert reply[0] == 0, reply
        pos = 1
        affected_rows, pos = _read_lcb(reply, pos)
        insert_id, pos = _read_lcb(reply, pos)
        server_status, nwarn = struct.unpack_from('<HH', reply, pos)
        pos += 4
        if nwarn:
            warnings.warn("Query {!r} caused {} warnings"
                .format(query, nwarn))
        return execute_result(insert_id, affected_rows)

    def query(self, query, *args, **kw):
        if args or kw:
            query = mysql_vformat(query, args, kw)
        chan = self.channel()
        buf = bytearray(b'\x00\x00\x00\x00\x03')
        buf += query.encode('utf-8')
        reply = chan.request(buf, QUERY).get()
        if reply[-1][0] == 0xff:
            _parse_error(reply[-1])
        assert reply[0][0] not in (0, 0xFF, 0xFE), \
            "Use execute for statements that does not return a result set"
        nfields, pos = _read_lcb(reply[0], 0)
        if pos < len(reply[0]):
            extra, pos = _read_lcb(reply[0], pos)
        else:
            extra = 0
        return Resultset(reply, nfields, extra)

    def _prepare(self, chan, query):
        buf = bytearray(b'\x00\x00\x00\x00\x16')
        buf += query.encode('utf-8')
        reply = chan.request(buf, PREPARED_STMT).get()
        if reply[-1][0] == 0xff:
            _parse_error(reply[-1])
        stmt_id, ncols, nparams, nwarn \
            = struct.unpack_from('<xLHHxH', reply[0])
        if nwarn:
            warnings.warn("Query {!r} caused {} warnings"
                .format(query, nwarn))
        params = []
        for pack in reply[1:nparams+1]:
            params.append(Field.parse_packet(pack))
        if nparams:
            fstart = nparams+2
        else:
            fstart = 1
        fields = []
        for pack in reply[fstart:fstart+ncols]:
            fields.append(Field.parse_packet(pack))
        stmt = PreparedStatement(query, stmt_id, fields, params)
        self._prepared[query] = stmt
        return stmt

    def execute_prepared(self, query, *args):
        chan = self.channel()
        stmt = self._prepared.get(query, None)
        if stmt is None:
            stmt = self._prepare(chan, query)

        if len(args) != len(stmt.params):
            raise TypeError("Expected {} parameters got {}".format(
                len(stmt.params), len(args)))
        buf = bytearray(b'\x00\x00\x00\x00\x17')
        buf += struct.pack('<L5x', stmt.id)
        la = len(args)
        for i in range(0, la, 8):
            byte = 0
            for j in range(i, min(i+8, la)):
                if args[j] is None:
                    byte |= 1 << (j & 7)
            buf.append(byte)
        if not stmt.bound:
            stmt.write_binding(buf)
        else:
            buf += b'\x00'
        for f, a in zip(stmt.params, args):
            if a is None:
                continue  # masked out
            if not isinstance(a, bytes):
                a = str(a).encode('utf-8')
            _write_lcbytes(buf, a)
        reply = chan.request(buf, QUERY).get()
        if reply[-1][0] == 0xff:
            _parse_error(reply[-1])
        res = self._parse_execute(reply, query)
        stmt.bound = True
        return res

    def query_prepared(self, query, *args):
        chan = self.channel()
        stmt = self._prepared.get(query, None)
        if stmt is None:
            stmt = self._prepare(chan, query)

        if len(args) != len(stmt.params):
            raise TypeError("Wrong number of parameters to prepared statement")
        buf = bytearray(b'\x00\x00\x00\x00\x17')
        buf += struct.pack('<L5x', stmt.id)
        la = len(args)
        for i in range(0, la, 8):
            byte = 0
            for j in range(i, min(i+8, la)):
                if args[j] is None:
                    byte |= 1 << j
            buf.append(byte)
        if not stmt.bound:
            stmt.write_binding(buf)
        else:
            buf += b'\x00'
        for a in args:
            if a is None:
                continue  # masked out
            if not isinstance(a, bytes):
                a = str(a).encode('utf-8')
            _write_lcbytes(buf, a)
        reply = chan.request(buf, QUERY).get()
        if reply[-1][0] == 0xff:
            _parse_error(reply[-1])
        assert reply[0][0] not in (0, 0xFF, 0xFE), \
            "Use execute for statements that does not return a result set"
        nfields, pos = _read_lcb(reply[0], 0)
        if pos < len(reply[0]):
            extra, pos = _read_lcb(reply[0], pos)
        else:
            extra = 0
        stmt.bound = True
        return BinaryResultset(stmt.row_class, reply, nfields, extra)





from functools import partial

from greenlet import getcurrent, GreenletExit

from .core import gethub, Condition
from . import sleep, TimeoutError


class Pool(object):

    def __init__(self, callback, *, limit, timeout):
        self.callback = callback
        self.limit = limit
        self.timeout = timeout
        self.current = 0
        self._cond = Condition()

    def __call__(self, *args, **kw):
        self.current += 1
        cur = getcurrent()
        killer = gethub().do_spawn(partial(self._timeout, cur))
        try:
            return self.callback(*args, **kw)
        finally:
            if not killer.dead:
                killer.detach().parent = cur
                killer.throw(GreenletExit())
            self.current -= 1
            self._cond.notify()

    def _timeout(self, task):
        sleep(self.timeout)
        cur = getcurrent()
        cur.parent = task.detach()
        raise TimeoutError()

    def wait_slot(self):
        while self.current >= self.limit:
            self._cond.wait()




from functools import wraps
from contextlib import contextmanager

from .core import Hub, gethub, Future, Condition, Lock, TimeoutError


__version__ = '0.2.a0'

__all__ = [
    'Hub',
    'gethub',
    'sleep',
    'Future',
    'Condition',
    'Lock',
    ]


def sleep(value):
    gethub().do_sleep(value)



import socket
import errno
from urllib.parse import urlencode

from .core import gethub, Lock
from . import channel
from .util import setcloexec

try:
    import ssl
except ImportError:
    pass


class Response(object):

    def __init__(self, status, headers, body):
        self.status = status
        self.headers = headers
        self.body = body


class RequestChannel(channel.PipelinedReqChannel):
    BUFSIZE = 4096

    def __init__(self, host, port, unixsock):
        super().__init__()
        if unixsock is None:
            self._sock = socket.socket(socket.AF_INET,
                socket.SOCK_STREAM, socket.IPPROTO_TCP)
            addr = (host, port)
        else:
            self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            addr = unixsock
        setcloexec(self._sock)
        self._sock.setblocking(0)
        self._connect(addr)
        self._start()

    def _connect(self, addr):
        try:
            self._sock.connect(addr)
        except socket.error as e:
            if e.errno == errno.EINPROGRESS:
                gethub().do_write(self._sock)
            else:
                raise

    def _close_channel(self):
        self._sock.close()
        super()._close_channel()

    def sender(self):
        buf = bytearray()

        add_chunk = buf.extend
        wait_write = gethub().do_write

        while True:
            if not buf:
                self.wait_requests()
            if not self._alive:
                return
            wait_write(self._sock)
            for chunk in self.get_pending_requests():
                add_chunk(chunk)
            try:
                bytes = self._sock.send(buf)
            except socket.error as e:
                if e.errno in (errno.EAGAIN, errno.EINTR):
                    continue
                else:
                    raise
            if not bytes:
                raise EOFError("Connection closed by peer")
            del buf[:bytes]

    def _readmore(self, buf, pos):
        while True:
            try:
                if pos[0]*2 > len(buf):
                    del buf[:pos[0]]
                    pos[0] = 0
                bytes = self._sock.recv(self.BUFSIZE)
                if not bytes:
                    raise EOFError("Connection closed by peer")
                buf.extend(bytes)
            except socket.error as e:
                if e.errno in (errno.EAGAIN, errno.EINTR):
                    gethub().do_read(self._sock)
                    continue
                else:
                    raise
            else:
                break

    def receiver(self):
        buf = bytearray()

        sock = self._sock
        pos = [0]

        def readchunked():
            result = bytearray()
            while True:
                while True:
                    idx = buf.find(b'\r\n', pos[0])
                    if idx >= 0:
                        break
                    self._readmore(buf, pos)
                line = buf[pos[0]:idx]
                pos[0] = idx
                try:
                    num = int(line, 16)
                except ValueError:
                    raise EOFError("Wrong number of bytes, or some extension")
                if num == 0:
                    break
                pos[0] += 2  # eat endline
                while len(buf) < pos[0] + num:
                    self._readmore(buf, pos)
                result.extend(buf[pos[0]:pos[0]+num])
                pos[0] += num
                if buf[pos[0]:pos[0]+2] != b'\r\n':
                    raise EOFError("Chunk is not terminated properly")
                pos[0] += 2
            while True:
                idx = buf.find(b'\r\n\r\n', pos[0])
                if idx >= 0:
                    break
                self._readmore(buf, pos)
            pos[0] = idx + 4  # Can't be any headers,
                              # but let's ignore them anyway
            return result

        def readrequest():
            while True:
                idx = buf.find(b'\r\n\r\n', pos[0])
                if idx >= 0:
                    break
                self._readmore(buf, pos)
            head = buf[pos[0]:idx]
            pos[0] = idx + 4
            lines = iter(head.decode('ascii').split('\r\n'))
            status = next(lines)
            headers = {}
            last_header = None
            for line in lines:
                if line.startswith((' ', '\t')):
                    if last_header is not None:
                        headers[last_header] += line
                    else:
                        raise EOFError("Wrong http headers")
                elif ':' in line:
                    k, v = line.split(':', 1)
                    k = k.strip()
                    if k in headers:
                        headers[k] += ', ' + v.strip()
                    else:
                        headers[k] = v.strip()
                else:
                    raise EOFError("Wrong http headers")
            te = headers.get('Transfer-Encoding', '')
            if te == 'chunked':
                return status, headers, readchunked()
            elif te:
                raise EOFError('Wrong transfer encoding {!r}'.format(te))

            clen = headers.get('Content-Length', None)

            if clen is None:
                if headers.get('Connection', '').lower() != 'close':
                    raise EOFError('Impossible to determine content length')
                try:
                    while True:
                        self._readmore(buf, pos)
                except EOFError:
                     return status, headers, buf

            clen = int(clen)
            if clen < 0:
                raise EOFError("Wrong content length")
            while pos[0] + clen > len(buf):
                self._readmore(buf, pos)
            return status, headers, buf[pos[0]:pos[0]+clen]

        while True:
            self.produce(readrequest())


class SecureRequestChannel(RequestChannel):

    def _connect(self, addr):
        super()._connect(addr)
        self._sock = ssl.SSLSocket(sock=self._sock,
            do_handshake_on_connect=False,
            ssl_version=ssl.PROTOCOL_TLSv1)
        while True:
            try:
                self._sock.do_handshake()
            except ssl.SSLWantReadError:
                gethub().do_read(self._sock)
            except ssl.SSLWantWriteError:
                gethub().do_write(self._sock)
            else:
                break

    def sender(self):
        buf = bytearray()

        add_chunk = buf.extend
        wait_write = gethub().do_write
        wait_read = gethub().do_read

        while True:
            if not buf:
                self.wait_requests()
            if not self._alive:
                return
            for chunk in self.get_pending_requests():
                add_chunk(chunk)
            try:
                bytes = self._sock.send(buf)
            except ssl.SSLWantReadError:
                wait_read(self._sock)
            except ssl.SSLWantWriteError:
                wait_write(self._sock)
            if not bytes:
                raise EOFError("Connection closed by peer")
            del buf[:bytes]

    def _readmore(self, buf, pos):
        while True:
            try:
                if pos[0]*2 > len(buf):
                    del buf[:pos[0]]
                    pos[0] = 0
                bytes = self._sock.recv(self.BUFSIZE)
                if not bytes:
                    raise EOFError("Connection closed by peer")
                buf.extend(bytes)
            except ssl.SSLWantReadError:
                gethub().do_read(self._sock)
            except ssl.SSLWantWriteError:
                gethub().do_write(self._sock)
            else:
                break


class HTTPClient(object):

    def __init__(self, host, port=80, unixsock=None, response_class=Response):
        self.host = host
        self.port = port
        self.unixsock = unixsock
        self.response_class = response_class
        self._channel = None
        self._channel_lock = Lock()

    def connection(self):
        if not self._channel:
            with self._channel_lock:
                if not self._channel:
                    self._channel = RequestChannel(self.host, self.port,
                        unixsock=self.unixsock)
        return self._channel

    def request(self, uri, *,
            method='GET',
            query=None,
            headers={},
            body=None):
        conn = self.connection()
        assert method.isidentifier(), method
        assert uri.startswith('/'), uri
        if query:
            if '?' in uri:
                uri += '&' + urlencode(query)
            else:
                uri += '?' + urlencode(query)
        headers = headers.copy()
        statusline = '{} {} HTTP/1.1'.format(method.upper(), uri)
        lines = [statusline]
        if isinstance(body, dict):
            body = urlencode(body)
        if isinstance(body, str):
            body = body.encode('utf-8')  # there are no other encodings, right?
        if body is not None:
            clen = len(body)
        else:
            clen = 0
            body = b''
        headers['Content-Length'] = clen
        for k, v in headers.items():
            lines.append('{}: {}'.format(k, str(v)))
        lines.append('')
        lines.append('')
        buf = '\r\n'.join(lines).encode('ascii')
        return self.response_class(*conn.request(buf + body).get())


class HTTPSClient(HTTPClient):

    def __init__(self, host, port=443, unixsock=None, response_class=Response):
        self.host = host
        self.port = port
        self.unixsock = unixsock
        self.response_class = response_class
        self._channel = None
        self._channel_lock = Lock()

    def connection(self):
        if not self._channel:
            with self._channel_lock:
                if not self._channel:
                    self._channel = SecureRequestChannel(self.host, self.port,
                        unixsock=self.unixsock)
        return self._channel



import socket
import errno
from time import time as current_time

from .core import gethub, Lock, Future
from .channel import PipelinedReqChannel


class Unix(PipelinedReqChannel):
    BUFSIZE = 1024

    def __init__(self, unixsock='/var/run/collectd-unixsock'):
        self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._sock.setblocking(0)
        self._cur_producing = []
        self._todo = 0
        try:
            self._sock.connect(unixsock)
        except socket.error as e:
            if e.errno == errno.EINPROGRESS:
                gethub().do_write(self._sock)
            else:
                raise
        super().__init__()
        self._start()

    def produce(self, line):
        if self._todo:
            self._cur_producing.append(line)
            self._todo -= 1
        else:
            num, tail = line.split(b' ', 1)
            lines = int(num)
            self._todo = lines
            self._cur_producing.append(line)
        if not self._todo:
            res = tuple(self._cur_producing)
            del self._cur_producing[:]
            self._producing.popleft()[1].set(res)

    def sender(self):
        buf = bytearray()

        add_chunk = buf.extend
        wait_write = gethub().do_write
        sock = self._sock

        while True:
            if not buf:
                self.wait_requests()
            wait_write(sock)
            for chunk in self.get_pending_requests():
                add_chunk(chunk)
            try:
                bytes = sock.send(buf)
            except socket.error as e:
                if e.errno in (errno.EAGAIN, errno.EINTR):
                    continue
                elif e.errno in (errno.EPIPE, errno.ECONNRESET):
                    raise EOFError()
                else:
                    raise
            if not bytes:
                raise EOFError()
            del buf[:bytes]

    def receiver(self):
        buf = bytearray()

        sock = self._sock
        wait_read = gethub().do_read
        add_chunk = buf.extend
        pos = 0
        current = []

        while True:
            if pos*2 > len(buf):
                del buf[:pos]
                pos = 0
            wait_read(sock)
            try:
                bytes = sock.recv(self.BUFSIZE)
                if not bytes:
                    raise EOFError()
                add_chunk(bytes)
            except socket.error as e:
                if e.errno in (errno.EAGAIN, errno.EINTR):
                    continue
                elif e.errno in (errno.EPIPE, errno.ECONNRESET):
                    raise EOFError()
                else:
                    raise
            while True:
                idx = buf.find(b'\n', pos)
                if idx < 0:
                    break
                line = buf[pos:idx]
                self.produce(line)
                pos = idx + 1


class Connection(object):

    def __init__(self, unixsock='/var/run/collectd-unixsock'):
        # TODO(tailhook)
        self.unixsock = unixsock
        self._channel = None
        self._channel_lock = Lock()

    def channel(self):
        if not self._channel:
            with self._channel_lock:
                if not self._channel:
                    self._channel = Unix(unixsock=self.unixsock)
        return self._channel

    def putval(self, identifier, values, interval=None, time=None):
        return self.putval_future(identifier, values,
                                  interval=interval, time=time).get()

    def putval_future(self, identifier, values, interval=None, time=None):
        buf = bytearray(b'PUTVAL ')
        buf += identifier.encode('ascii')
        if interval is not None:
            buf += ' interval={0:d}'.format(interval).encode('ascii')
        for tup in values:
            if time is None:
                time = current_time()
            lst = [str(int(time))]
            for val in tup:
                if val is None:
                    lst.append('U')
                else:
                    lst.append(str(float(val)))
            buf += b' '
            buf += ':'.join(lst).encode('ascii')
        return self.channel().request(buf)

    def flush(self, timeout=None, plugin=None, identifier=None):
        buf = bytearray(b'FLUSH')
        if timeout is not None:
            buf += ' timeout={}'.format(timeout).encode('ascii')
        if plugin is not None:
            buf += ' plugin={}'.format(plugin).encode('ascii')
        if identifier is not None:
            buf += ' identifier={}'.format(identifier).encode('ascii')
        return self.channel().request(buf).get()

    def putnotif(self, message, severity='warning', time=None, host=None,
        plugin=None, plugin_instance=None, type=None, type_instance=None):
        buf = bytearray(b'PUTNOTIF')
        buf += ' message="{}"'.format(message.replace('"',"'")).encode('ascii')
        buf += ' severity={}'.format(severity).encode('ascii')
        if time is None:
            time = int(current_time())
        buf += ' time={}'.format(time).encode('ascii')
        if host:
            ' host={}'.format(host).encode('ascii')
        if plugin:
            ' plugin={}'.format(plugin).encode('ascii')
        if plugin_instance:
            ' plugin_instance={}'.format(plugin_instance).encode('ascii')
        if type:
            ' type={}'.format(type).encode('ascii')
        if type_instance:
            ' type_instance={}'.format(type_instance).encode('ascii')
        return self.channel().request(buf).get()

    def getval(self, identifier):
        buf = bytearray(b'GETVAL ')
        buf += identifier.encode('ascii')
        res = {}
        for line in self.channel().request(buf).get()[1:]:
            name, value = line.split(b'=')
            res[name.decode('ascii')] = float(value)
        return res

    def listval(self):
        buf = bytearray(b'LISTVAL')
        for line in self.channel().request(buf).get()[1:]:
            num, ident = line.split(b' ')
            num = int(num)
            yield ident.decode('ascii'), num

    # TODO(tailhook) implement other commands




import heapq
import time
import threading
import select
import weakref
import logging
import errno
from collections import deque, defaultdict
from functools import partial
from operator import methodcaller
from math import ceil

import greenlet

from .util import priorityqueue, orderedset, socket_pair, marker_object

__all__ = [
    'Zorrolet',
    'Hub',
    'gethub',
    'Future',
    'Condition',
    'Lock',
    ]

FUTURE_EXCEPTION = marker_object('FUTURE_EXCEPTION')
FUTURE_PENDING = marker_object('FUTURE_PENDING')

os_errors = (IOError, OSError)


class TimeoutError(Exception):
    pass


class WaitingError(Exception):
    """Is raised when poll returned POLLHUP or POLLERR"""


class Zorrolet(greenlet.greenlet):
    __slots__ = ('hub', 'cleanup')

    def __init__(self, fun, hub):
        super().__init__(fun, parent=hub._self)
        self.hub = hub
        self.cleanup = []

    def detach(self):
        for cleanup in self.cleanup:
            cleanup(self)
        del self.cleanup[:]
        return self

    def __repr__(self):
        if self.gr_frame:
            return '<Z{:x} {}:{}>'.format(id(self),
                self.gr_frame.f_code.co_filename, self.gr_frame.f_lineno)
        elif self.dead:
            return '<Z{:x} dead>'.format(id(self))
        else:
            r = getattr(self, 'run', None)
            if isinstance(r, partial):
                return '<Z{:x} {}>'.format(id(self),
                    getattr(r.func, '__name__', r.func))
            else:
                return '<Z{:x} {}>'.format(id(self),
                    getattr(r, '__name__', r))


class EpollWrapper(object):
    delegate_methods = ('register', 'unregister', 'modify', 'close')

    def __init__(self, poll):
        self._poll = poll
        for name in self.delegate_methods:
            setattr(self, name, getattr(poll, name))

    def poll(self, timeout=-1, maxevents=-1):
        if timeout > 0:
            return self._poll.poll(timeout/1000, maxevents)
        else:
            return self._poll.poll(timeout, maxevents)


class Hub(object):
    def __init__(self):
        self._log = logging.getLogger('zorro.hub.{:x}'.format(id(self)))
        self._queue = orderedset()
        try:
            self._poller = EpollWrapper(select.epoll())
            self.POLLIN = select.EPOLLIN
            self.POLLOUT = select.EPOLLOUT
            self.POLLHUP = select.EPOLLHUP
            self.POLLERR = select.EPOLLERR
        except AttributeError:
            self._poller = select.poll()
            self.POLLIN = select.POLLIN
            self.POLLOUT = select.POLLOUT
            self.POLLHUP = select.POLLHUP
            self.POLLERR = select.POLLERR
            self._log.info("Using poller %r", self._poller)
        self._filedes = methodcaller('fileno')
        self._timeouts = priorityqueue()
        self._in_sockets = defaultdict(list)
        self._out_sockets = defaultdict(list)
        self._control = socket_pair()
        self._poller.register(self._control[0], self.POLLIN)
        self._control_fd = self._control[0].fileno()
        self._start_tasks = []
        self._services = weakref.WeakKeyDictionary()
        self._tasks = weakref.WeakKeyDictionary()
        self._helpers = weakref.WeakKeyDictionary()

    # global methods

    def change_poller(self, cls, *, POLLIN, POLLOUT, POLLERR, POLLHUP,
        filedes=methodcaller('fileno')):
        self._log.info("Changing poller from %r to %r",
            self._poller.__class__, cls)
        if hasattr(self._poller, 'close'):
            self._poller.close()
        self._poller = cls()
        self.POLLIN = POLLIN
        self.POLLOUT = POLLOUT
        self.POLLHUP = POLLHUP
        self.POLLERR = POLLERR
        self._filedes = filedes
        for k in set(self._in_sockets).union(self._out_sockets):
            msk = 0
            if k in self._in_sockets:
                msk |= POLLIN
            elif k in self._out_sockets:
                msk |= POLLOUT
            self._poller.register(k, msk)
        if hasattr(self, '_control'):
            self._poller.register(self._control[0], self.POLLIN)
            self._control_fd = self._control[0].fileno()

    def wakeup(self):
        self._control[1].send(b'x')

    def stop(self):
        """Stop all services, and wait for other tasks to complete"""
        self._log.warning("Stop called from thread ``%s'' and %r",
            threading.current_thread().name, greenlet.getcurrent())
        self.stopping = True
        if threading.current_thread().ident != self._thread:
            self.wakeup()
        else:
            self.shutdown_tasks(self._services, self._tasks)

    def shutdown_tasks(self, src, tgt):
        for i in list(src.keys()):
            del src[i]
            tgt[i] = None
            try:
                i.detach().throw(greenlet.GreenletExit())
            except BaseException as e:
                self.log_exception(e)

    def crash(self):
        """Rude stop of hub at next iteration"""
        self._log.warning("Crash called from thread ``%s'' and %r",
            threading.current_thread().name, greenlet.getcurrent())
        self.stopped = True
        if threading.current_thread().ident != self._thread:
            self.wakeup()

    def run(self, *tasks):
        self.stopping = False
        self.stopped = False
        self._self = greenlet.getcurrent()
        self._thread = threading.current_thread().ident
        self._log.warning("Starting in thread ``%s'' and %r",
            threading.current_thread().name, self._self)
        for f in self._start_tasks:
            self.do_spawn(f)
        del self._start_tasks
        for f in tasks:
            self.do_spawn(f)
        while True:
            self.queue()
            self.timeouts()

            if self.stopped:
                self._log.warning("Breaking main loop")
                break
            elif self.stopping and self._services:
                self._log.warning("Stopping services")
                self.shutdown_tasks(self._services, self._tasks)
            elif not self._tasks and not self._services and self._helpers:
                self._log.warning("No more active tasks, stopping helpers")
                self.shutdown_tasks(self._helpers, self._tasks)
            if not self._tasks and not self._services:
                break

            if not self._queue:
                self.io()

        self._log.warning("Hub stopped")
        self.stopping = True
        self.stopped = True

    # Logging methods

    def log_plugged(self, plugin, name):
        self._log.info("Plugged in %r under the name %r", plugin, name)

    def log_exception(self, e):
        self._log.error("Exception in one of spawned tasks", exc_info=e)

    # Internals

    def queue(self):
        while not self.stopped:
            tsk = self._queue.first()
            if tsk is None:
                return
            t = tsk[0]
            t.detach()
            try:
                t.switch(*tsk[1:])
            except BaseException as e:
                self.log_exception(e)

    def io(self):
        timeo = self._timeouts.min()
        if timeo is not None:
            timeo = int(ceil(max(timeo - time.time(), 0)*1000))
        else:
            timeo = -1
        try:
            items = self._poller.poll(timeo)
        except os_errors as e:
            if e.errno in (errno.EINTR, errno.EAGAIN):
                return
            raise
        else:
            POLLOUT = self.POLLOUT
            POLLIN = self.POLLIN
            POLLERR = self.POLLHUP | self.POLLERR
            for fd, ev in items:
                if ev & POLLOUT:
                    if fd in self._out_sockets:
                        task = self._out_sockets[fd][0]
                        self.queue_task(task, 'read')
                if ev & POLLIN:
                    if fd in self._in_sockets:
                        task = self._in_sockets[fd][0]
                        self.queue_task(task, 'write')
                    elif fd == self._control_fd:
                        self._control[0].recv(1024)
                        # throw it, just need wake up
                if ev & POLLERR:
                    if fd in self._in_sockets:
                        task = self._in_sockets[fd][0]
                        self.queue_task(task, 'err')
                    if fd in self._out_sockets:
                        task = self._out_sockets[fd][0]
                        self.queue_task(task, 'err')


    def timeouts(self):
        if not self._timeouts:
            return
        now = time.time()
        while True:
            task = self._timeouts.pop(now)
            if task is None:
                break
            self.queue_task(task, 'timeout')

    def queue_task(self, task, *value):
        task.detach()
        task.cleanup.append(self._queue.add(task, *value))
        return task

    # Helper methods
    def do_sleep(self, tm, more=False):
        targ = time.time() + tm
        let = greenlet.getcurrent()
        let.cleanup.append(self._timeouts.add(targ, let))
        del let # no cycles
        if not more:
            self._self.switch()

    def _check_mask(self, fd, new=False):
        msk = 0
        if fd in self._in_sockets:
            msk |= self.POLLIN
        if fd in self._out_sockets:
            msk |= self.POLLOUT
        if new:
            self._poller.register(fd, msk)
        elif msk:
            self._poller.modify(fd, msk)
        else:
            self._poller.unregister(fd)

    def _queue_sock(self, sock, dic, odic, more=False):
        fd = self._filedes(sock)
        def deque_sock(let):
            items.remove(let)
            if not items:
                del dic[fd]
            self._check_mask(fd)
        let = greenlet.getcurrent()
        let.cleanup.append(deque_sock)
        items = dic[fd]
        items.append(let)
        self._check_mask(fd, new=len(items) == 1 and fd not in odic)
        del let
        if not more:
            val = self._self.switch()
            if val == 'err':
                raise WaitingError()

    def do_read(self, sock, more=False):
        self._queue_sock(sock, self._in_sockets, self._out_sockets, more=more)

    def do_write(self, sock, more=False):
        self._queue_sock(sock, self._out_sockets, self._in_sockets, more=more)

    def do_spawnservice(self, fun):
        if self.stopping:
            raise RuntimeError("Trying to spawn a service while stopping")
        let = Zorrolet(fun, self)
        self._services[let] = None
        self.queue_task(let)
        return let

    def do_spawn(self, fun):
        let = Zorrolet(fun, self)
        self._tasks[let] = None
        self.queue_task(let)
        return let

    def do_spawnswitch(self, fun):
        let = Zorrolet(fun, self)
        self._tasks[let] = None
        self.queue_task(greenlet.getcurrent())
        let.switch()
        return let

    def do_spawnhelper(self, fun):
        let = Zorrolet(fun, self)
        self._helpers[let] = None
        self.queue_task(let)
        return let

    def add_task(self, fun):
        self._start_tasks.append(fun)


def gethub():
    let = greenlet.getcurrent()
    return let.hub


class Condition(object):

    def __init__(self):
        self._queue = deque()

    def notify(self):
        if self._queue:
            tsk = self._queue[0]
            tsk.hub.queue_task(tsk)

    def wait(self, timeout=None):
        cur = greenlet.getcurrent()
        cur.cleanup.append(self._queue.remove)
        self._queue.append(cur)
        hub = cur.hub
        if timeout is not None:
            targ = time.time() + timeout
            cur.cleanup.append(hub._timeouts.add(targ, cur))
        del cur # no cycles
        hub._self.switch()


class Future(object):
    def __init__(self, fun=None):
        self._listeners = []
        self._value = FUTURE_PENDING
        if fun is not None:
            def future():
                try:
                    result = fun()
                except Exception as e:
                    self.throw(e)
                else:
                    self.set(result)
            gethub().do_spawn(future)

    def get(self, timeout=None):
        val = self._value
        if val is not FUTURE_PENDING:
            if val is FUTURE_EXCEPTION:
                raise self._exception
            else:
                return val
        cur = greenlet.getcurrent()
        cur.cleanup.append(self._listeners.remove)
        self._listeners.append(cur)
        hub = cur.hub
        if timeout is not None:
            targ = time.time() + timeout
            cur.cleanup.append(hub._timeouts.add(targ, cur))
        del cur # no cycles
        hub._self.switch()
        val = self._value
        if val is FUTURE_PENDING:
            raise TimeoutError()
        if val is FUTURE_EXCEPTION:
            raise self._exception
        else:
            return val

    def set(self, value):
        if self._value is not FUTURE_PENDING:
            raise RuntimeError("Value is already set")
        self._value = value
        lst = self._listeners
        del self._listeners
        hub = gethub()
        for one in lst:
            hub.queue_task(one)

    def throw(self, exception):
        self._value = FUTURE_EXCEPTION
        self._exception = exception
        lst = self._listeners
        del self._listeners
        for one in lst:
            gethub().queue_task(one)

    def check(self):
        return self._value is FUTURE_PENDING


class Lock(Condition):
    def __init__(self):
        super().__init__()
        self._locked = False

    def acquire(self):
        while self._locked:
            self.wait()
        self._locked = True

    def release(self):
        self._locked = False
        self.notify()

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.release()



from distutils.core import setup


setup(name='Zorro',
      version='0.1.4',
      description='Network communication library, with zeromq support',
      author='Paul Colomiets',
      author_email='paul@colomiets.name',
      url='http://github.com/tailhook/zorro',
      classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        ],
      packages=['zorro', 'zorro.mongodb'],
    )
